"""Search the AF database for different objects."""

import abc
import warnings
from collections.abc import Iterator, Sequence
from typing import Generic, TypeVar

from . import Asset, EventFrame, dotnet

SearchResultType = TypeVar(
    "SearchResultType",
    # PIAFBase.PIAFElement,
    # PIAFBase.PIAFEventFrame,
    # PIAFAttribute.PIAFAttribute,
)
AFSearchResultType = TypeVar("AFSearchResultType", covariant=True)


class SearchResult(Generic[AFSearchResultType, SearchResultType], abc.ABC):
    """Container for search results."""

    def __init__(
        self,
        search: "dotnet.AF.Search.AFSearch[AFSearchResultType]",
    ) -> None:
        self.search = search
        self.result_type: type[SearchResultType]

    @property
    def name(self) -> str:
        """Return the name of the search result."""
        return self.search.SearchName

    @property
    def count(self) -> int:
        """Return the number of items in the search result."""
        return self.search.GetTotalCount()

    def __iter__(self) -> Iterator[SearchResultType]:
        """Return an iterator over the items in the search result."""
        for item in self.search.FindObjects():
            yield self.result_type(item)  # type: ignore

    def one(self) -> SearchResultType:
        """Return the only item in the search result.

        Raises
        ------
            ValueError: If there are no results or more than one result.
        """
        if self.count == 0:
            raise ValueError("No results found")
        if self.count > 1:
            raise ValueError(f"More than one result found ({self.count} results)")
        return next(iter(self))

    @abc.abstractmethod
    def to_list(self) -> Sequence[SearchResultType]:
        """Return all items in the search result in a sequence."""
        pass


class AttributeSearchResult(SearchResult[dotnet.AF.Asset.AFAttribute, Asset.AFAttribute]):
    """Container for attribute search results."""

    def __init__(
        self,
        search: dotnet.AF.Search.AFAttributeSearch,
    ) -> None:
        super().__init__(search)
        self.result_type = Asset.AFAttribute

    def to_list(self) -> Asset.AFAttributeList:
        """Return all items in the search result."""
        return Asset.AFAttributeList(list(self))


class ElementSearchResult(SearchResult[dotnet.AF.Asset.AFElement, Asset.AFElement]):
    """Container for attribute search results."""

    def __init__(
        self,
        search: dotnet.AF.Search.AFElementSearch,
    ) -> None:
        super().__init__(search)
        self.result_type = Asset.AFElement

    def to_list(self) -> Asset.AFElementList:
        """Return all items in the search result."""
        return Asset.AFElementList(list(self))


class EventFrameSearchResult(
    SearchResult[dotnet.AF.EventFrame.AFEventFrame, EventFrame.AFEventFrame]
):
    """Container for attribute search results."""

    def __init__(
        self,
        search: dotnet.AF.Search.AFEventFrameSearch,
    ) -> None:
        super().__init__(search)
        self.result_type = EventFrame.AFEventFrame

    def to_list(self) -> EventFrame.AFEventFrameList:
        """Return all items in the search result."""
        return EventFrame.AFEventFrameList(list(self))


class Search:
    """Search the AF database for different objects."""

    def __init__(self, database: dotnet.AF.AFDatabase) -> None:
        self.database = database

    def attributes(
        self, query: str, query_name: str = "attribute_search"
    ) -> AttributeSearchResult:
        """Search for elements in the AF database."""
        search = dotnet.lib.AF.Search.AFAttributeSearch(self.database, query_name, query)
        return AttributeSearchResult(search)

    def elements(self, query: str, query_name: str = "element_search") -> ElementSearchResult:
        """Search for elements in the AF database."""
        search = dotnet.lib.AF.Search.AFElementSearch(self.database, query_name, query)
        return ElementSearchResult(search)

    def event_frames(
        self, query: str, query_name: str = "event_frame_search"
    ) -> EventFrameSearchResult:
        """Search for event frames in the AF database."""
        search = dotnet.lib.AF.Search.AFEventFrameSearch(self.database, query_name, query)
        return EventFrameSearchResult(search)

    def _descendant(self, path: str) -> Asset.AFElement:
        return Asset.AFElement(self.database.Elements.get_Item(path))

    def __call__(self, query: str | list[str]) -> Asset.AFAttributeList:
        """Search AFAttributes by element|attribute path strings.

        Return a list of AFAttributes directly from a list of element|attribute path strings

            like this:

        list("BaseElement/childElement/childElement|Attribute|ChildAttribute|ChildAttribute",
        "BaseElement/childElement/childElement|Attribute|ChildAttribute|ChildAttribute")

        """
        warnings.warn(
            """Call to Search.__call__ is deprecated, use Search.<element_type> instead""",
            DeprecationWarning,
            stacklevel=2,
        )
        attributelist = Asset.AFAttributeList([])
        if isinstance(query, list):
            for x in query:
                attributelist.extend(self(x))
        elif "|" in query:
            splitpath = query.split("|")
            elem = self._descendant(splitpath[0])
            attribute = elem.attributes[splitpath[1]]
            if len(splitpath) > 2:
                for x in range(len(splitpath) - 2):
                    attribute = attribute.children[splitpath[x + 2]]
            attributelist.append(attribute)
        return attributelist



"""Type hints for the AF.Data module.

Contains various enumerations and classes for data retrieval.
"""

import enum

from . import Generic, Time
from . import UnitsOfMeasure as UOM
from ._values import AFValue, AFValues


class AFBoundaryType(enum.IntEnum):
    """Mock class of the AF.Data.AFBoundaryType enumeration."""

    Inside = 0
    Outside = 1
    Interpolated = 2


class AFBufferOption(enum.IntEnum):
    """Mock class of the AF.Data.AFBufferOption enumeration."""

    DoNotBuffer = 0
    BufferIfPossible = 1
    Buffer = 2


class AFCalculationBasis(enum.IntEnum):
    """Mock class of the AF.Data.AFCalculationBasis enumeration."""

    TimeWeighted = 0
    EventWeighted = 1
    TimeWeightedContinuous = 2
    TimeWeightedDiscrete = 3
    EventWeightedExcludeMostRecentEvent = 4
    EventWeightedExcludeEarliestEvent = 5
    EventWeightedIncludeBothEnds = 6


class AFRetrievalMode(enum.IntEnum):
    Auto = 0
    AtOrBefore = 1
    Before = 6
    AtOrAfter = 2
    After = 7
    Exact = 4


class AFSampleType(enum.IntEnum):
    """Mock class of the AF.Data.AFSampleType enumeration."""

    ExpressionRecordedValues = 0
    Interval = 1


class AFSummaryTypes(enum.IntEnum):
    """Mock class of the AF.Data.AFSummaryTypes enumeration."""

    None_ = 0
    Total = 1
    Average = 2
    Minimum = 4
    Maximum = 8
    Range = 16
    StdDev = 32
    PopulationStdDev = 64
    Count = 128
    PercentGood = 8192
    TotalWithUOM = 16384
    All = 24831
    AllForNonNumeric = 8320


class AFTimestampCalculation(enum.IntEnum):
    """Mock class of the AF.Data.AFTimestampCalculation enumeration."""

    Auto = 0
    EarliestTime = 1
    MostRecentTime = 2


class AFUpdateOption(enum.IntEnum):
    """Mock class of the AF.Data.AFUpdateOption enumeration."""

    Replace = 0
    Insert = 4
    NoReplace = 2
    ReplaceOnly = 3
    InsertNoCompression = 5
    Remove = 6


SummariesDict = Generic.Dictionary[AFSummaryTypes, AFValues]
SummaryDict = Generic.Dictionary[AFSummaryTypes, AFValue]


class AFData:
    """Mock class of the AF.Data.AFData class."""

    @staticmethod
    def FilteredSummaries(
        time_range: Time.AFTimeRange,
        interval: Time.AFTimeSpan,
        filter_expression: str,
        summary_types: AFSummaryTypes,
        calculation_basis: AFCalculationBasis,
        filter_evaluation: AFSampleType,
        filter_interval: Time.AFTimeSpan,
        time_type: AFTimestampCalculation,
        /,
    ) -> SummariesDict:
        return SummariesDict([])

    @staticmethod
    def InterpolatedValue(
        time: Time.AFTime,
        uom: UOM.UOM,
        /,
    ) -> AFValue:
        return AFValue(None, time)

    @staticmethod
    def InterpolatedValues(
        time_range: Time.AFTimeRange,
        interval: Time.AFTimeSpan,
        uom: UOM.UOM,
        filter_expression: str,
        include_filtered_values: bool,
        /,
    ) -> AFValues:
        return AFValues()

    @staticmethod
    def RecordedValue(
        time: Time.AFTime,
        retrieval_mode: AFRetrievalMode,
        uom: UOM.UOM,
        /,
    ) -> AFValue:
        return AFValue(None, time)

    @staticmethod
    def RecordedValues(
        time_range: Time.AFTimeRange,
        boundary_type: AFBoundaryType,
        uom: UOM.UOM,
        filter_expression: str,
        include_filtered_values: bool,
        /,
    ) -> AFValues:
        return AFValues()

    @staticmethod
    def Summaries(
        time_range: Time.AFTimeRange,
        interval: Time.AFTimeSpan,
        summary_type: AFSummaryTypes,
        calculation_basis: AFCalculationBasis,
        time_type: AFTimestampCalculation,
        /,
    ) -> SummariesDict:
        return SummariesDict([])

    @staticmethod
    def Summary(
        time_range: Time.AFTimeRange,
        summary_type: AFSummaryTypes,
        calculation_basis: AFCalculationBasis,
        time_type: AFTimestampCalculation,
        /,
    ) -> SummaryDict:
        return SummaryDict([])

    @staticmethod
    def UpdateValue(
        value: AFValue,
        update_option: AFUpdateOption,
        buffer_option: AFBufferOption,
        /,
    ) -> None:
        pass



"""PIconnect - Connector to the OSISoft PI and PI-AF databases."""

from PIconnect.config import PIConfig  # noqa: I001 isort: skip
from PIconnect.AF import AFDatabase, PIAFDatabase
from PIconnect.dotnet import lib, load_SDK
from PIconnect.PI import PIServer

from . import _version


def __getattr__(name: str):
    """Lazy load the AF SDK."""
    match name:
        case "__sdk_version":
            return tuple(int(x) for x in lib.AF_SDK_VERSION.split("_")[0].split("."))
        case _:
            raise AttributeError(f"module {__name__} has no attribute {name}")


__version__ = _version.get_versions()["version"]

__all__ = [
    "AFDatabase",
    "PIAFDatabase",
    "PIConfig",
    "PIServer",
    "load_SDK",
]



"""Mock classes of the AF.PI namespace of the OSIsoft PI-AF SDK."""

import enum
from collections.abc import Iterable, Iterator

from PIconnect._typing import System

from . import Data, Generic, Time, _values

__all__ = ["PIPoint", "PIServer", "PIServers"]


class PIConnectionInfo:
    """Mock class of the AF.PI.PIConnectionInfo class."""

    def __init__(self) -> None:
        self.OperationTimeOut: System.TimeSpan


class PIAuthenticationMode(enum.IntEnum):
    """Mock class of the AF.PI.PIAuthenticationMode class."""

    WindowsAuthentication = 0
    PIUserAuthentication = 1


class PIServer:
    """Mock class of the AF.PI.PIServer class."""

    def __init__(self, name: str) -> None:
        self.ConnectionInfo = PIConnectionInfo()
        self.Name = name
        self._connected = False

    def Connect(
        self,
        retry: bool | System.Net.NetworkCredential,
        authentication_mode: PIAuthenticationMode | None = None,
    ) -> None:
        """Stub for connecting to test server."""
        self._connected = True

    def Disconnect(self) -> None:
        """Stub for disconnecting from test server."""
        self._connected = False


class PIServers:
    """Mock class of the AF.PI.PIServers class."""

    def __init__(self) -> None:
        self.DefaultPIServer = PIServer("Testing")

    def __iter__(self) -> Iterator[PIServer]:
        return (x for x in [self.DefaultPIServer])

    def __getitem__(self, server: str) -> PIServer | None:
        """Stub to mock getting a server by name."""
        if server == self.DefaultPIServer.Name:
            return self.DefaultPIServer


class PIPoint:
    """Mock class of the AF.PI.PIPoint class."""

    Name: str = "TestPIPoint"
    """This property identifies the name of the PIPoint"""

    Step: bool = False
    """This property indicates whether the PIPoint is a step PIPoint"""

    @staticmethod
    def CurrentValue() -> _values.AFValue:
        return _values.AFValue(None)

    @staticmethod
    def FilteredSummaries(
        time_range: Time.AFTimeRange,
        interval: Time.AFTimeSpan,
        filter_expression: str,
        summary_type: Data.AFSummaryTypes,
        calculation_basis: Data.AFCalculationBasis,
        sample_type: Data.AFSampleType,
        sample_interval: Time.AFTimeSpan,
        time_type: Data.AFTimestampCalculation,
        /,
    ) -> Data.SummariesDict:
        return Data.SummariesDict([])

    @staticmethod
    def FindPIPoints(
        connection: PIServer,
        query: str,
        source: str | None,
        attribute_names: Iterable[str] | None,
    ) -> Iterable["PIPoint"]:
        """Stub to mock querying PIPoints."""
        return []

    @staticmethod
    def GetAttributes(names: list[str], /) -> Generic.PropertyDict:
        return Generic.PropertyDict([])

    @staticmethod
    def InterpolatedValue(time: Time.AFTime, /) -> _values.AFValue:
        return _values.AFValue(None, time)

    @staticmethod
    def InterpolatedValues(
        time_range: Time.AFTimeRange,
        interval: Time.AFTimeSpan,
        filter_expression: str,
        include_filtered_values: bool,
        /,
    ) -> _values.AFValues:
        return _values.AFValues()

    @staticmethod
    def LoadAttributes(params: list[str], /) -> None:
        pass

    @staticmethod
    def RecordedValue(
        time: Time.AFTime, retrieval_mode: Data.AFRetrievalMode, /
    ) -> _values.AFValue:
        return _values.AFValue(None, time)

    @staticmethod
    def RecordedValues(
        time_range: Time.AFTimeRange,
        boundary_type: Data.AFBoundaryType,
        filter_expression: str,
        include_filtered_values: bool,
        max_count: int = 0,
        /,
    ) -> _values.AFValues:
        return _values.AFValues()

    @staticmethod
    def Summaries(
        time_range: Time.AFTimeRange,
        interval: Time.AFTimeSpan,
        summary_type: Data.AFSummaryTypes,
        calculation_basis: Data.AFCalculationBasis,
        time_type: Data.AFTimestampCalculation,
        /,
    ) -> Data.SummariesDict:
        return Data.SummariesDict([])

    @staticmethod
    def Summary(
        time_range: Time.AFTimeRange,
        summary_type: Data.AFSummaryTypes,
        calculation_basis: Data.AFCalculationBasis,
        time_type: Data.AFTimestampCalculation,
    ) -> Data.SummaryDict:
        return Data.SummaryDict([])

    @staticmethod
    def UpdateValue(
        value: _values.AFValue,
        update_mode: Data.AFUpdateOption,
        buffer_option: Data.AFBufferOption,
        /,
    ) -> None:
        pass



"""PI - Core containers for connections to PI databases."""

import enum
import warnings
from typing import Any, cast

import PIconnect._typing.AF as _AFtyping
from PIconnect import Data, Time, dotnet

__all__ = ["PIServer", "PIPoint"]


class InitialisationWarning(UserWarning):
    pass


class AuthenticationMode(enum.IntEnum):
    """AuthenticationMode indicates how a user authenticates to a PI Server.

    Detailed information is available at
    :afsdk:`AF.PI.PIAuthenticationMode <T_OSIsoft_AF_PI_PIAuthenticationMode.htm>`.
    """

    #: Use Windows authentication when making a connection
    WINDOWS_AUTHENTICATION = 0
    #: Use the PI User authentication mode when making a connection
    PI_USER_AUTHENTICATION = 1


_DEFAULT_AUTH_MODE = AuthenticationMode.PI_USER_AUTHENTICATION


def _lookup_servers() -> dict[str, dotnet.AF.PI.PIServer]:
    servers: dict[str, dotnet.AF.PI.PIServer] = {}

    for server in dotnet.lib.AF.PI.PIServers():
        try:
            servers[server.Name] = server
        except (Exception, dotnet.lib.System.Exception) as e:  # type: ignore
            warnings.warn(
                f"Failed loading server data for {server.Name} "
                f"with error {type(cast(Exception, e)).__qualname__}",
                InitialisationWarning,
                stacklevel=2,
            )
    return servers


def _lookup_default_server() -> dotnet.AF.PI.PIServer | None:
    default_server = None
    try:
        default_server = dotnet.lib.AF.PI.PIServers().DefaultPIServer
    except Exception:
        warnings.warn("Could not load the default PI Server", ResourceWarning, stacklevel=2)
    return default_server


class PIPoint(Data.DataContainer):
    """Reference to a PI Point to get data and corresponding metadata from the server.

    Parameters
    ----------
        pi_point (AF.PI.PIPoint): Reference to a PIPoint as returned by the SDK
    """

    version = "0.3.0"

    def __init__(self, pi_point: dotnet.AF.PI.PIPoint) -> None:
        super().__init__()
        self.pi_point = pi_point
        self.tag = pi_point.Name
        self.__attributes_loaded = False
        self.__raw_attributes = {}

    def __repr__(self):
        """Return the string representation of the PI Point."""
        return (
            f"{self.__class__.__qualname__}({self.tag}, {self.description}; "
            f"Current Value: {self.current_value} {self.units_of_measurement})"
        )

    @property
    def created(self):
        """Return the creation datetime of a point."""
        return Time.timestamp_to_index(self.raw_attributes["creationdate"])

    @property
    def description(self):
        """Return the description of the PI Point.

        .. todo::

            Add setter to alter displayed description
        """
        return self.raw_attributes["descriptor"]

    @property
    def last_update(self):
        """Return the time at which the last value for this PI Point was recorded."""
        return Time.timestamp_to_index(self.pi_point.CurrentValue().Timestamp.UtcTime)

    @property
    def name(self) -> str:
        """Return the name of the PI Point."""
        return self.tag

    @property
    def raw_attributes(self) -> dict[str, Any]:
        """Return a dictionary of the raw attributes of the PI Point."""
        self.__load_attributes()
        return self.__raw_attributes

    @property
    def units_of_measurement(self) -> str | None:
        """Return the units of measument in which values for this PI Point are reported."""
        return self.raw_attributes["engunits"]

    @property
    def stepped_data(self) -> bool:
        """Return False when the PIPoint contains continuous data or True when stepped data."""
        return self.pi_point.Step

    def __load_attributes(self) -> None:
        """Load the raw attributes of the PI Point from the server."""
        if not self.__attributes_loaded:
            self.pi_point.LoadAttributes([])
            self.__attributes_loaded = True
        self.__raw_attributes = {att.Key: att.Value for att in self.pi_point.GetAttributes([])}

    def _current_value(self) -> Any:
        """Return the last recorded value for this PI Point (internal use only)."""
        return self.pi_point.CurrentValue().Value

    def _filtered_summaries(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        interval: dotnet.AF.Time.AFTimeSpan,
        filter_expression: str,
        summary_types: dotnet.AF.Data.AFSummaryTypes,
        calculation_basis: dotnet.AF.Data.AFCalculationBasis,
        filter_evaluation: dotnet.AF.Data.AFSampleType,
        filter_interval: dotnet.AF.Time.AFTimeSpan,
        time_type: dotnet.AF.Data.AFTimestampCalculation,
    ) -> _AFtyping.Data.SummariesDict:
        return self.pi_point.FilteredSummaries(
            time_range,
            interval,
            filter_expression,
            summary_types,
            calculation_basis,
            filter_evaluation,
            filter_interval,
            time_type,
        )

    def _interpolated_value(self, time: dotnet.AF.Time.AFTime) -> dotnet.AF.Asset.AFValue:
        """Return a single value for this PI Point."""
        return self.pi_point.InterpolatedValue(time)

    def _interpolated_values(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        interval: dotnet.AF.Time.AFTimeSpan,
        filter_expression: str,
    ) -> dotnet.AF.Asset.AFValues:
        include_filtered_values = False
        return self.pi_point.InterpolatedValues(
            time_range, interval, filter_expression, include_filtered_values
        )

    def _normalize_filter_expression(self, filter_expression: str) -> str:
        return filter_expression.replace("%tag%", self.tag)

    def _recorded_value(
        self, time: dotnet.AF.Time.AFTime, retrieval_mode: dotnet.AF.Data.AFRetrievalMode
    ) -> dotnet.AF.Asset.AFValue:
        """Return a single recorded value for this PI Point."""
        return self.pi_point.RecordedValue(
            time, dotnet.lib.AF.Data.AFRetrievalMode(int(retrieval_mode))
        )

    def _recorded_values(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        boundary_type: dotnet.AF.Data.AFBoundaryType,
        filter_expression: str,
    ) -> dotnet.AF.Asset.AFValues:
        include_filtered_values = False
        return self.pi_point.RecordedValues(
            time_range, boundary_type, filter_expression, include_filtered_values
        )

    def _summary(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        summary_types: dotnet.AF.Data.AFSummaryTypes,
        calculation_basis: dotnet.AF.Data.AFCalculationBasis,
        time_type: dotnet.AF.Data.AFTimestampCalculation,
    ) -> _AFtyping.Data.SummaryDict:
        return self.pi_point.Summary(time_range, summary_types, calculation_basis, time_type)

    def _summaries(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        interval: dotnet.AF.Time.AFTimeSpan,
        summary_types: dotnet.AF.Data.AFSummaryTypes,
        calculation_basis: dotnet.AF.Data.AFCalculationBasis,
        time_type: dotnet.AF.Data.AFTimestampCalculation,
    ) -> _AFtyping.Data.SummariesDict:
        return self.pi_point.Summaries(
            time_range, interval, summary_types, calculation_basis, time_type
        )

    def _update_value(
        self,
        value: dotnet.AF.Asset.AFValue,
        update_mode: dotnet.AF.Data.AFUpdateOption,
        buffer_mode: dotnet.AF.Data.AFBufferOption,
    ) -> None:
        return self.pi_point.UpdateValue(value, update_mode, buffer_mode)


class PIServer(object):  # pylint: disable=useless-object-inheritance
    """PIServer is a connection to an OSIsoft PI Server.

    Parameters
    ----------
        server (str, optional): Name of the server to connect to, defaults to None
        username (str, optional): can be used only with password as well
        password (str, optional): -//-
        todo: domain, auth
        timeout (int, optional): the maximum seconds an operation can take

    .. note::
        If the specified `server` is unknown a warning is thrown and the connection
        is redirected to the default server, as if no server was passed. The list
        of known servers is available in the `PIServer.servers` dictionary.
    """

    version = "0.2.2"

    #: Dictionary of known servers, as reported by the SDK
    _servers: dict[str, dotnet.AF.PI.PIServer] | None = None
    _default_server: dotnet.AF.PI.PIServer | None = None

    @classmethod
    def servers(cls) -> dict[str, dotnet.AF.PI.PIServer]:
        """Return a dictionary of the known servers."""
        if cls._servers is None:
            cls._servers = _lookup_servers()
        return cls._servers

    @classmethod
    def default_server(cls) -> dotnet.AF.PI.PIServer | None:
        """Return the default server."""
        if cls._default_server is None:
            cls._default_server = _lookup_default_server()
        return cls._default_server

    def __init__(
        self,
        server: str | None = None,
        username: str | None = None,
        password: str | None = None,
        domain: str | None = None,
        authentication_mode: AuthenticationMode = _DEFAULT_AUTH_MODE,
        timeout: int | None = None,
    ) -> None:
        default_server = self.default_server()
        if server is None:
            if default_server is None:
                raise ValueError("No server was specified and no default server was found.")
            self.connection = default_server
        else:
            if (_server := dotnet.lib.AF.PI.PIServers()[server]) is not None:
                self.connection = _server
            else:
                if default_server is None:
                    raise ValueError(
                        f"Server '{server}' not found and no default server was found."
                    ) from None
                message = 'Server "{server}" not found, using the default server.'
                warnings.warn(
                    message=message.format(server=server), category=UserWarning, stacklevel=1
                )
                self.connection = default_server

        if bool(username) != bool(password):
            raise ValueError(
                "When passing credentials both the username and password must be specified."
            )
        if domain and not username:
            raise ValueError(
                "A domain can only specified together with a username and password."
            )
        if username:
            secure_pass = dotnet.lib.System.Security.SecureString()
            if password is not None:
                for c in password:
                    secure_pass.AppendChar(c)
            cred = (username, secure_pass) + ((domain,) if domain else ())
            self._credentials = (
                dotnet.lib.System.Net.NetworkCredential(cred[0], cred[1], *cred[2:]),
                dotnet.lib.AF.PI.PIAuthenticationMode(int(authentication_mode)),
            )
        else:
            self._credentials = None

        if timeout:
            # System.TimeSpan(hours, minutes, seconds)
            self.connection.ConnectionInfo.OperationTimeOut = dotnet.lib.System.TimeSpan(
                0, 0, timeout
            )

    def __enter__(self):
        """Open connection context with the PI Server."""
        if self._credentials:
            self.connection.Connect(*self._credentials)
        else:
            # Don't force to retry connecting if previous attempt failed
            force_connection = False
            self.connection.Connect(force_connection)
        return self

    def __exit__(self, *args: Any):
        """Close connection context with the PI Server."""
        self.connection.Disconnect()

    def __repr__(self) -> str:
        """Representation of the PIServer object."""
        return f"{self.__class__.__qualname__}(\\\\{self.server_name})"

    @property
    def server_name(self):
        """Name of the connected server."""
        return self.connection.Name

    def search(self, query: str | list[str], source: str | None = None) -> list[PIPoint]:
        """Search PIPoints on the PIServer.

        Parameters
        ----------
            query (str or [str]): String or list of strings with queries
            source (str, optional): Defaults to None. Point source to limit the results

        Returns
        -------
            list: A list of :class:`PIPoint` objects as a result of the query

        .. todo::

            Reject searches while not connected
        """
        if isinstance(query, list):
            return [y for x in query for y in self.search(x, source)]
        # elif not isinstance(query, str):
        #     raise TypeError('Argument query must be either a string or a list of strings,' +
        #                     'got type ' + str(type(query)))
        return [
            PIPoint(pi_point)
            for pi_point in dotnet.lib.AF.PI.PIPoint.FindPIPoints(
                self.connection, str(query), source, None
            )
        ]



"""Fake classes to mask SDK complexity."""

import dataclasses
import datetime
from collections.abc import Iterable
from typing import Any, Generic, TypeVar

import pytest
import pytz

import PIconnect._typing.AF as AF
import PIconnect.PI as PI


@dataclasses.dataclass
class _UTCTime:
    Year: int
    Month: int
    Day: int
    Hour: int
    Minute: int
    Second: int
    Millisecond: int


class FakeAFTime(object):
    """Fake AFTime to mask away SDK complexity."""

    def __init__(self, timestamp: datetime.datetime):
        self.UtcTime = _UTCTime(
            timestamp.year,
            timestamp.month,
            timestamp.day,
            timestamp.hour,
            timestamp.minute,
            timestamp.second,
            int(timestamp.microsecond / 1000),
        )


_a = TypeVar("_a")
_b = TypeVar("_b")


class FakeKeyValue(Generic[_a, _b]):
    """Container for fake Key:Value pairs."""

    def __init__(self, key: _a, value: _b) -> None:
        self.Key = key
        self.Value = value


class FakeAFValue(AF.Asset.AFValue, Generic[_a]):
    """Fake AFValue to mask away SDK complexity."""

    def __init__(self, value: _a, timestamp: datetime.datetime):
        self.Value = value
        self.Timestamp = FakeAFTime(timestamp)


class FakeAFValues(AF.Asset.AFValues, Generic[_a]):
    """Fake AFValues to mask away SDK complexity."""

    def __init__(self, values: Iterable[FakeAFValue[_a]]):
        self.Value = list(values)
        self.Count = len(self.Value)


class FakePIPoint_(Generic[_a]):
    """Fake PI Point to mask away SDK complexity."""

    def __init__(
        self,
        tag: str,
        values: Iterable[_a],
        timestamps: Iterable[datetime.datetime],
        attributes: dict[str, Any],
    ):
        self.Name = tag
        self.values = [
            FakeAFValue(value, timestamp)
            for value, timestamp in zip(values, timestamps, strict=True)
        ]
        self.attributes = [FakeKeyValue(*att) for att in attributes.items()]


class FakePIPoint(AF.PI.PIPoint, Generic[_a]):
    """Fake PI Point to mask away SDK complexity."""

    def __init__(self, pi_point: FakePIPoint_[_a]) -> None:
        self.pi_point = pi_point
        self.call_stack = ["%s created" % self.__class__.__name__]
        self.Name = pi_point.Name

    def CurrentValue(self) -> FakeAFValue[_a]:
        """Return the current value of the PI Point."""
        self.call_stack.append("CurrentValue called")
        return self.pi_point.values[-1]

    def LoadAttributes(self, *args: Any, **kwargs: Any) -> None:
        """Load the attributes of the PI Point."""
        self.call_stack.append("LoadAttributes called")

    def GetAttributes(self, *args: Any, **kwargs: Any) -> list[FakeKeyValue[str, Any]]:
        """Return the attributes of the PI Point."""
        self.call_stack.append("GetAttributes called")
        return self.pi_point.attributes

    def RecordedValues(self, *args: Any, **kwargs: Any) -> list[FakeAFValue[_a]]:
        """Return the recorded values of the PI Point."""
        self.call_stack.append("RecordedValues called")
        return self.pi_point.values

    def InterpolatedValues(self, *args: Any, **kwargs: Any) -> list[FakeAFValue[_a]]:
        """Return the interpolated values of the PI Point."""
        self.call_stack.append("InterpolatedValues called")
        return self.pi_point.values


class VirtualTestCase(object):
    """Test VirtualPIPoint addition."""

    def __init__(self):
        self.tag = "TEST_140_053_FQIS053_01_Meetwaarde"
        self.values = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.timestamp_numbers = [
            1502654535.813,
            1502671554.038,
            1502695584.315,
            1502704569.874,
            1502709576.898,
            1502713512.168,
            1502718534.453,
            1502722585.816,
            1502731598.316,
            1502732545.013,
        ]
        self.timestamps = [
            datetime.datetime.fromtimestamp(x, tz=pytz.utc) for x in self.timestamp_numbers
        ]
        self.attributes = {"engunits": "m3/h", "descriptor": "Flow"}
        pi_point = FakePIPoint_(
            tag=self.tag,
            values=self.values,
            timestamps=self.timestamps,
            attributes=self.attributes,
        )
        self.point = PI.PIPoint(FakePIPoint(pi_point))


@pytest.fixture
def pi_point() -> VirtualTestCase:
    """Return a VirtualTestCase object."""
    return VirtualTestCase()



"""Mocking the Generic types from System.Collections.Generic namespace.

TODO: Migrate to the `_typing.dotnet` module.
"""

from collections.abc import Iterable, Iterator
from typing import Any, Generic, TypeVar

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class DictItem(Generic[_KT, _VT]):
    def __init__(self, key: _KT, value: _VT) -> None:
        self.Key = key
        self.Value = value


class Dictionary(Generic[_KT, _VT]):
    def __init__(self, items: Iterable[tuple[_KT, _VT]]) -> None:
        self.Items = items

    def __iter__(self) -> Iterator[DictItem[_KT, _VT]]:
        for item in self.Items:
            yield DictItem(*item)


class TimeSpan:
    def __init__(self, hours: int, minutes: int, seconds: int) -> None:
        self.Hours = hours
        self.Minutes = minutes
        self.Seconds = seconds


class SecureString:
    def __init__(self) -> None:
        self.Value = ""

    def AppendChar(self, char: str) -> None:
        self.Value += char


class NetworkCredential:
    def __init__(
        self, username: str, password: SecureString, domain: str | None = None
    ) -> None:
        self.UserName = username
        self.Password = password
        self.Domain = domain


PropertyDict = Dictionary[str, Any]



"""Test communication with the PI System."""

import datetime

import pytest
import pytz

import PIconnect as PI
import PIconnect.PI as PI_
from PIconnect import dotnet

from .fakes import VirtualTestCase, pi_point

dotnet.lib.load_test_SDK()

__all__ = ["TestServer", "TestSearchPIPoints", "TestPIPoint", "pi_point"]


class TestServer:
    """Test connecting to the server."""

    def test_connection(self):
        """Test that creating a PI.PIServer object without arguments raises no exception."""
        PI.PIServer()

    def test_server_name(self):
        """Test that the server reports the same name as which was connected to."""
        default_server = PI.PIServer.default_server()
        if default_server is None:
            pytest.skip("No default server found.")
        else:
            servername = default_server.Name
            server = PI.PIServer(servername)
            assert server.server_name == servername

    def test_warn_unkown_server(self):
        """Test that the server reports a warning when an unknown host is specified."""
        server_names = list(PI.PIServer.servers())
        server_name = "__".join(server_names + ["UnknownHostName"])
        with pytest.warns(UserWarning):
            PI.PIServer(server_name)

    def test_repr(self):
        """Test that the server representation matches the connected server."""
        default_server = PI.PIServer.default_server()
        if default_server is None:
            pytest.skip("No default server found.")
        else:
            servername = default_server.Name
            server = PI.PIServer(servername)
            assert repr(server) == "PIServer(\\\\{})".format(servername)


class TestSearchPIPoints:
    """Test searching for PI Points on the default server."""

    def test_search_single_string(self):
        """Test searching for PI points using a single string."""
        with PI.PIServer() as server:
            points = server.search("L_140_053*")
            assert isinstance(points, list)
            for point in points:
                assert isinstance(point, PI_.PIPoint)

    def test_search_multiple_strings(self):
        """Tests searching for PI points using a list of strings."""
        with PI.PIServer() as server:
            points = server.search(["L_140_053*", "M_127*"])
            assert isinstance(points, list)
            for point in points:
                assert isinstance(point, PI_.PIPoint)

    # def test_search_integer_raises_error(self):
    #     """Tests searching for PI points using an integer raises a TypeError."""
    #     with PI.PIServer() as server, self.assertRaises(TypeError):
    #         server.search(1)


class TestPIPoint:
    """Test valid interface of PIPoint."""

    def test_repr(self, pi_point: VirtualTestCase):
        """Test representation of the PI Point."""
        assert repr(pi_point.point) == "%s(%s, %s; Current Value: %s %s)" % (
            "PIPoint",
            pi_point.tag,
            pi_point.attributes["descriptor"],
            pi_point.values[-1],
            pi_point.attributes["engunits"],
        )

    def test_name(self, pi_point: VirtualTestCase):
        """Test retrieving the name of the PI Point."""
        assert pi_point.point.tag == pi_point.tag

    def test_current_value(self, pi_point: VirtualTestCase):
        """Test retrieving the current value from a PI point."""
        assert pi_point.point.current_value == pi_point.values[-1]

    def test_last_update(self, pi_point: VirtualTestCase):
        """Test retrieving the last update timestamp."""
        origin = datetime.datetime(1970, 1, 1).replace(tzinfo=pytz.utc)
        assert (
            round(
                (pi_point.point.last_update - origin).total_seconds()
                - pi_point.timestamp_numbers[-1],
                ndigits=7,
            )
            == 0
        )

    def test_units_of_measurement(self, pi_point: VirtualTestCase):
        """Test retrieving the units of measurement of the returned PI point."""
        assert pi_point.point.units_of_measurement == pi_point.attributes["engunits"]

    def test_description(self, pi_point: VirtualTestCase):
        """Test retrieving the description of the PI point."""
        assert pi_point.point.description == pi_point.attributes["descriptor"]

    def test_raw_attributes(self, pi_point: VirtualTestCase):
        """Test retrieving the attributes of the PI point as a dict."""
        assert pi_point.point.raw_attributes == pi_point.attributes

    def test_recorded_values_values(self, pi_point: VirtualTestCase):
        """Test retrieving some recorded data from the server."""
        data = pi_point.point.recorded_values("01-07-2017", "02-07-2017")
        assert list(data.values) == pi_point.values

    def test_recorded_values_timestamps(self, pi_point: VirtualTestCase):
        """Test retrieving some recorded data from the server."""
        data = pi_point.point.recorded_values("01-07-2017", "02-07-2017")
        assert list(data.index) == pi_point.timestamps

    def test_interpolated_values_values(self, pi_point: VirtualTestCase):
        """Test retrieving some interpolated data from the server."""
        data = pi_point.point.interpolated_values("01-07-2017", "02-07-2017", "1h")
        assert list(data.values) == pi_point.values

    def test_interpolated_values_timestamps(self, pi_point: VirtualTestCase):
        """Test retrieving some interpolated data from the server."""
        data = pi_point.point.interpolated_values("01-07-2017", "02-07-2017", "1h")
        assert list(data.index) == pi_point.timestamps



"""AFSDK - Loads the .NET libraries from the OSIsoft AF SDK."""

import logging
import os
import pathlib
import sys
from typing import cast

from ._typing import AF, AF_SDK_VERSION, AFType, System, SystemType

__all__ = ["AF", "System", "AF_SDK_VERSION", "lib", "load_SDK"]

logger = logging.getLogger(__name__)

StrPath = str | pathlib.Path


class dotNET:
    """Class to load the .NET libraries from the OSIsoft AF SDK."""

    def __init__(self) -> None:
        self._af: AFType | None = None
        self._system: SystemType | None = None
        self._af_sdk_version: str | None = None

    @property
    def AF(self) -> AFType:
        """Return the AF SDK."""
        if self._af is None:
            raise ImportError(".NET libraries not loaded, call PIconnect.load_SDK() first.")
        return self._af

    @property
    def System(self) -> SystemType:
        """Return the System SDK."""
        if self._system is None:
            raise ImportError(".NET libraries not loaded, call PIconnect.load_SDK() first.")
        return self._system

    @property
    def AF_SDK_VERSION(self) -> str:
        """Return the AF SDK version."""
        return self.AF.PISystems().Version

    def load(self, assembly_path: StrPath | None = None) -> None:
        """Return a new instance of the PI connector."""
        full_path = _get_SDK_path(assembly_path)
        if full_path is None:
            if assembly_path:
                raise ImportError(f"AF SDK not found at '{assembly_path}'")
            raise ImportError(
                "AF SDK not found, check installation "
                "or pass valid path to directory containing SDK assembly."
            )
        self._af, self._system = _get_dotnet_libraries(full_path)
        self._af_sdk_version = self.AF.PISystems().Version
        logger.info("Loaded AF SDK version %s", self._af_sdk_version)

    def load_test_SDK(self) -> None:
        self._af = AF
        self._system = System
        self._af_sdk_version = AF_SDK_VERSION


def _get_dotnet_libraries(full_path: StrPath) -> tuple[AFType, SystemType]:
    import clr  # type: ignore

    sys.path.append(str(full_path))
    clr.AddReference("OSIsoft.AFSDK")  # type: ignore ; pylint: disable=no-member
    import System  # type: ignore
    from OSIsoft import AF  # type: ignore

    _AF = cast(AFType, AF)
    _System = cast(SystemType, System)
    return _AF, _System


def _get_SDK_path(full_path: StrPath | None = None) -> pathlib.Path | None:
    if full_path:
        assembly_directories = [pathlib.Path(full_path)]
    else:
        installation_directories = {
            os.getenv("PIHOME"),
            "C:\\Program Files\\PIPC",
            "C:\\Program Files (x86)\\PIPC",
        }
        assembly_directories = (
            pathlib.Path(path) / "AF\\PublicAssemblies\\4.0\\"
            for path in installation_directories
            if path is not None
        )
    for AF_dir in assembly_directories:
        logging.debug("Full path to potential SDK location: '%s'", AF_dir)
        if AF_dir.is_dir():
            return AF_dir


lib = dotNET()


def load_SDK(assembly_path: StrPath | None = None) -> None:
    """Load the AF SDK from the specified path.

    Parameters
    ----------
        assembly_path (str | Path, optional): Path to the AF SDK assembly. If None, the default
            installation path will be used.

    Raises
    ------
        ImportError: If the AF SDK cannot be found or loaded.
    """
    global lib
    lib.load(assembly_path)



"""Mock classes for the System.Security module."""


class SecureString:
    def __init__(self) -> None:
        self.Value = ""

    def AppendChar(self, char: str) -> None:
        self.Value += char



"""Configuration for PIconnect package."""


class PIConfigContainer:
    """Configuration for PIconnect package.

    This should not be instantiated directly, but accessed through the `PIConfig` object.
    """

    _default_timezone: str = ""

    def __init__(self) -> None:
        self.DEFAULT_TIMEZONE = "UTC"

    @property
    def DEFAULT_TIMEZONE(self) -> str:
        """Timezone in which values are returned."""
        return self._default_timezone

    @DEFAULT_TIMEZONE.setter
    def DEFAULT_TIMEZONE(self, value: str) -> None:
        import zoneinfo

        if value not in zoneinfo.available_timezones():
            raise ValueError("{v!r} not found in pytz.all_timezones".format(v=value))
        self._default_timezone = value


PIConfig = PIConfigContainer()



"""AF - Core containers for connections to the PI Asset Framework."""

import logging
import warnings
from typing import Any, Self

from PIconnect import Asset, EventFrame, Search, Time, dotnet

_logger = logging.getLogger(__name__)
_DEFAULT_EVENTFRAME_SEARCH_MODE = EventFrame.EventFrameSearchMode.STARTING_AFTER


class AFDatabase:
    """Context manager for connections to the PI Asset Framework database."""

    version = "0.3.0"

    @classmethod
    def servers(cls) -> dict[str, dotnet.AF.PISystem]:
        """Return a dictionary of the known servers."""
        return {server.Name: server for server in dotnet.lib.AF.PISystems()}

    @classmethod
    def default_server(cls) -> dotnet.AF.PISystem | None:
        """Return the default server."""
        if dotnet.lib.AF.PISystems().DefaultPISystem:
            return dotnet.lib.AF.PISystems().DefaultPISystem
        servers = dotnet.lib.AF.PISystems()
        if servers.Count > 0:
            return next(iter(servers))
        else:
            return None

    def __init__(self, server: str | None = None, database: str | None = None) -> None:
        self.server = self._initialise_server(server)
        self.database = self._initialise_database(database)
        self.search = Search.Search(self.database)

    def _initialise_server(self, server: str | None) -> dotnet.AF.PISystem:
        """Initialise the server connection."""
        _logger.debug(f"Initialising server connection from {server!r}")
        default_server = self.default_server()
        if server is None:
            if default_server is None:
                raise ValueError("No server specified and no default server found.")
            _logger.debug(f"Using default server: {default_server.Name}")
            return default_server

        if (_server := dotnet.lib.AF.PISystems()[server]) is not None:
            _logger.debug(_server)
            return _server
        else:
            if default_server is None:
                raise ValueError(
                    f'Server "{server}" not found and no default server found.'
                ) from None
            message = f'Server "{server}" not found, using the default server.'
            _logger.debug(message)
            warnings.warn(message=message, category=UserWarning, stacklevel=2)
            return default_server

    def _initialise_database(self, database: str | None) -> dotnet.AF.AFDatabase:
        def default_db():
            default = self.server.Databases.DefaultDatabase
            if default is None:
                raise ValueError("No database specified and no default database found.")
            return default

        if database is None:
            return default_db()

        if (_db := self.server.Databases[database]) is not None:
            _logger.debug(_db)
            return _db
        else:
            message = f'Database "{database}" not found, using the default database.'
            warnings.warn(message=message, category=UserWarning, stacklevel=2)
            return default_db()

    def __enter__(self) -> Self:
        """Open the PI AF server connection context."""
        self.server.Connect()
        return self

    def __exit__(
        self,
        *args: Any,  # type: ignore
    ) -> bool:
        """Close the PI AF server connection context."""
        _logger.log(0, f"Closing connection to {self} ({args=})")
        return False
        # Disabled disconnecting because garbage collection sometimes impedes
        # connecting to another server later
        # self.server.Disconnect()

    def __repr__(self) -> str:
        """Return a representation of the PI AF database connection."""
        return f"{self.__class__.__qualname__}(\\\\{self.server_name}\\{self.database_name})"

    @property
    def server_name(self) -> str:
        """Return the name of the connected PI AF server."""
        return self.server.Name

    @property
    def database_name(self) -> str:
        """Return the name of the connected PI AF database."""
        return self.database.Name

    @property
    def children(self) -> dict[str, Asset.AFElement]:
        """Return a dictionary of the direct child elements of the database."""
        return {c.Name: Asset.AFElement(c) for c in self.database.Elements}

    @property
    def tables(self) -> dict[str, Asset.AFTable]:
        """Return a dictionary of the tables in the database."""
        return {t.Name: Asset.AFTable(t) for t in self.database.Tables}

    def descendant(self, path: str) -> Asset.AFElement:
        """Return a descendant of the database from an exact path."""
        return Asset.AFElement(self.database.Elements.get_Item(path))

    def event_frames(
        self,
        start_time: Time.TimeLike = "",
        start_index: int = 0,
        max_count: int = 1000,
        search_mode: EventFrame.EventFrameSearchMode = _DEFAULT_EVENTFRAME_SEARCH_MODE,
        search_full_hierarchy: bool = False,
    ) -> dict[str, EventFrame.AFEventFrame]:
        """Search for event frames in the database."""
        _start_time = Time.to_af_time(start_time)
        _search_mode = dotnet.lib.AF.EventFrame.AFEventFrameSearchMode(int(search_mode))
        return {
            frame.Name: EventFrame.AFEventFrame(frame)
            for frame in dotnet.lib.AF.EventFrame.AFEventFrame.FindEventFrames(
                self.database,
                None,
                _start_time,
                start_index,
                max_count,
                _search_mode,
                None,
                None,
                None,
                None,
                search_full_hierarchy,
            )
        }


class PIAFDatabase(AFDatabase):
    """Context manager for connections to the PI Asset Framework database."""

    version = "0.3.0"

    def __init__(self, server: str | None = None, database: str | None = None) -> None:
        warnings.warn(
            "PIAFDatabase is deprecated, use AFDatabase instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(server=server, database=database)



"""Enumerations for PI options."""

import enum


class UpdateMode(enum.IntEnum):
    """Indicates how to treat duplicate values in the archive.

    Only used when supported by the Data Reference.

    Detailed information is available at
    :afsdk:`AF.Data.AFUpdateOption <T_OSIsoft_AF_Data_AFUpdateOption.htm>`
    """

    #: Add the value to the archive.
    #: If any values exist at the same time, will overwrite one of them and set its
    #: Substituted flag.
    REPLACE = 0
    #: Add the value to the archive. Any existing values at the same time are not overwritten.
    INSERT = 1
    #: Add the value to the archive only if no value exists at the same time.
    #: If a value already exists for that time, the passed value is ignored.
    NO_REPLACE = 2
    #: Replace an existing value in the archive at the specified time.
    #: If no existing value is found, the passed value is ignored.
    REPLACE_ONLY = 3
    #: Add the value to the archive without compression.
    #: If this value is written to the snapshot, the previous snapshot value will be written to
    #: the archive,
    #: without regard to compression settings.
    #: Note that if a subsequent snapshot value is written without the InsertNoCompression
    #: option,
    #: the value added with the InsertNoCompression option is still subject to compression.
    INSERT_NO_COMPRESSION = 5
    #: Remove the value from the archive if a value exists at the passed time.
    REMOVE = 6


class BufferMode(enum.IntEnum):
    """Indicates buffering option in updating values, when supported by the Data Reference.

    Detailed information is available at
    :afsdk:`AF.Data.AFBufferOption <T_OSIsoft_AF_Data_AFBufferOption.htm>`
    """

    #: Updating data reference values without buffer.
    DO_NOT_BUFFER = 0
    #: Try updating data reference values with buffer.
    #: If fails (e.g. data reference AFDataMethods does not support Buffering,
    #: or its Buffering system is not available),
    #: then try updating directly without buffer.
    BUFFER_IF_POSSIBLE = 1
    # Updating data reference values with buffer.
    BUFFER = 2


class CalculationBasis(enum.IntEnum):
    """CalculationBasis indicates how values should be weighted over a time range.

    Detailed information is available at
    :afsdk:`AF.Data.AFCalculationBasis <T_OSIsoft_AF_Data_AFCalculationBasis.htm>`.
    """

    #: Each event is weighted according to the time over which it applies.
    TIME_WEIGHTED = 0
    #: Each event is weighted equally.
    EVENT_WEIGHTED = 1
    #: Each event is time weighted, but interpolation is always done as if it is
    #: continous data.
    TIME_WEIGHTED_CONTINUOUS = 2
    #: Each event is time weighted, but interpolation is always done as if it is
    #: discrete, stepped, data.
    TIME_WEIGHTED_DISCRETE = 3
    #: Each event is weighted equally, except data at the end of the interval is
    #: excluded.
    EVENT_WEIGHTED_EXCLUDE_MOST_RECENT = 4
    #: Each event is weighted equally, except data at the beginning of the interval
    #: is excluded.
    EVENT_WEIGHTED_EXCLUDE_EARLIEST = 5
    #: Each event is weighted equally, data at both boundaries of the interval are
    #: explicitly included.
    EVENT_WEIGHTED_INCLUDE_BOTH_ENDS = 6


class ExpressionSampleType(enum.IntEnum):
    """ExpressionSampleType indicates how expressions are evaluated over a time range.

    Detailed information is available at
    :afsdk:`AF.Data.AFSampleType <T_OSIsoft_AF_Data_AFSampleType.htm>`.
    """

    #: The expression is evaluated at each archive event.
    EXPRESSION_RECORDED_VALUES = 0
    #: The expression is evaluated at a sampling interval, passed as a separate argument.
    INTERVAL = 1


class RetrievalMode(enum.IntEnum):
    """RetrievalMode indicates which recorded value should be returned.

    Detailed information is available at
    :afsdk:`AF.Data.AFRetrievalMode <T_OSIsoft_AF_Data_AFRetrievalMode.htm>`.
    """

    #: Autmatic detection
    AUTO = 0
    #: At the exact time if available, else the first before the requested time
    AT_OR_BEFORE = 1
    #: The first before the requested time
    BEFORE = 6
    #: At the exact time if available, else the first after the requested time
    AT_OR_AFTER = 2
    #: The first after the requested time
    AFTER = 7
    #: At the exact time if available, else return an error
    EXACT = 4


class SummaryType(enum.IntFlag):
    """SummaryType indicates which types of summary should be calculated.

    Based on :class:`enum.IntEnum` in Python 3.5 or earlier. `SummaryType`'s can
    be or'ed together. Python 3.6 or higher returns a new `IntFlag`, while in
    previous versions it will be casted down to `int`.

    >>> SummaryType.MINIMUM | SummaryType.MAXIMUM  # Returns minimum and maximum
    <SummaryType.MAXIMUM|MINIMUM: 12>  # On Python 3.6+
    12  # On previous versions

    Detailed information is available at
    :afsdk:`AF.Data.AFSummaryTypes <T_OSIsoft_AF_Data_AFSummaryTypes.htm>`.
    """

    #: No summary data
    NONE = 0
    #: A total over the time span
    TOTAL = 1
    #: Average value over the time span
    AVERAGE = 2
    #: The minimum value in the time span
    MINIMUM = 4
    #: The maximum value in the time span
    MAXIMUM = 8
    #: The range of the values (max-min) in the time span
    RANGE = 16
    #: The sample standard deviation of the values over the time span
    STD_DEV = 32
    #: The population standard deviation of the values over the time span
    POP_STD_DEV = 64
    #: The sum of the event count (when the calculation is event weighted).
    #: The sum of the event time duration (when the calculation is time weighted.)
    COUNT = 128
    #: The percentage of the data with a good value over the time range.
    #: Based on time for time weighted calculations,
    #: based on event count for event weigthed calculations.
    PERCENT_GOOD = 8192
    #: The total over the time span,
    #: with the unit of measurement that's associated with the input
    #: (or no units if not defined for the input).
    TOTAL_WITH_UOM = 16384
    #: A convenience to retrieve all summary types
    ALL = 24831
    #: A convenience to retrieve all summary types for non-numeric data
    ALL_FOR_NON_NUMERIC = 8320


class TimestampCalculation(enum.IntEnum):
    """
    TimestampCalculation defines the timestamp returned for a given summary calculation.

    Detailed information is available at
    :afsdk:`AF.Data.AFTimeStampCalculation <T_OSIsoft_AF_Data_AFTimestampCalculation.htm>`.
    """

    #: The timestamp is the event time of the minimum or maximum for those summaries
    #: or the beginning of the interval otherwise.
    AUTO = 0
    #: The timestamp is always the beginning of the interval.
    EARLIEST_TIME = 1
    #: The timestamp is always the end of the interval.
    MOST_RECENT_TIME = 2


class EventFrameSearchMode(enum.IntEnum):
    """EventFrameSearchMode.

    EventFrameSearchMode defines the interpretation and direction from the start time
    when searching for event frames.

    Detailed information is available at
    :afsdk:`AF.EventFrame.AFEventFrameSearchMode <T_OSIsoft_AF_EventFrame_AFEventFrameSearchMode.htm>`.
    including a graphical display of event frames that are returned for a given search
    mode.
    """  # noqa: E501

    #: Uninitialized
    NONE = 0
    #: Backward from start time, also known as starting before
    BACKWARD_FROM_START_TIME = 1
    STARTING_BEFORE = 1
    #: Forward from start time, also known as starting after
    FORWARD_FROM_START_TIME = 2
    STARTING_AFTER = 2
    #: Backward from end time, also known as ending before
    BACKWARD_FROM_END_TIME = 3
    ENDING_BEFORE = 3
    #: Forward from end time, also known as ending after
    FORWARD_FROM_END_TIME = 4
    ENDING_AFTER = 4
    #: Backward in progress, also known as starting before and in progress
    BACKWARD_IN_PROGRESS = 5
    STARTING_BEFORE_IN_PROGRESS = 5
    #: Forward in progress, also known as starting after and in progress
    FORWARD_IN_PROGRESS = 6
    STARTING_AFTER_IN_PROGRESS = 6



"""Common fixtures for testing PIconnect."""

import os

import pytest


def on_CI() -> bool:
    """Return True if the tests are running on a CI environment."""
    return (
        os.getenv("GITHUB_ACTIONS", "false").lower() == "true"
        or os.getenv("TF_BUILD", "false").lower() == "true"
        or os.getenv("READTHEDOCS", "false").lower() == "true"
    )


skip_if_on_CI = pytest.mark.skipif(on_CI(), reason="Real SDK not available on CI")



"""Time related functions and classes."""

import datetime
import zoneinfo

import pandas as pd  # type: ignore

from PIconnect import dotnet
from PIconnect.config import PIConfig

TimeLike = str | datetime.datetime
IntervalLike = str | datetime.timedelta | pd.Timedelta


def to_af_time_range(start_time: TimeLike, end_time: TimeLike) -> dotnet.AF.Time.AFTimeRange:
    """Convert a combination of start and end time to a time range.

    Both `start_time` and `end_time` can be either a :any:`datetime.datetime` object or
    a string.
    `datetime` objects are first converted to a string, before being passed to
    :afsdk:`AF.Time.AFTimeRange <M_OSIsoft_AF_Time_AFTimeRange__ctor_1.htm>`.
    It is also possible to specify either end as a `datetime` object,
    and then specify the other boundary as a relative string.

    Parameters
    ----------
        start_time (str | datetime): Start time of the time range.
        end_time (str | datetime): End time of the time range.

    Returns
    -------
        :afsdk:`AF.Time.AFTimeRange <T_OSIsoft_AF_Time_AFTimeRange.htm>`:
            Time range covered by the start and end time.
    """
    if isinstance(start_time, datetime.datetime):
        start_time = start_time.isoformat()
    if isinstance(end_time, datetime.datetime):
        end_time = end_time.isoformat()

    return dotnet.lib.AF.Time.AFTimeRange.Parse(start_time, end_time)


def to_af_time(time: TimeLike) -> dotnet.AF.Time.AFTime:
    """Convert a time to a AFTime value.

    Parameters
    ----------
        time (str | datetime): Time to convert to AFTime.

    Returns
    -------
        :afsdk:`AF.Time.AFTime <T_OSIsoft_AF_Time_AFTime.htm>`:
            AFTime version of time.
    """
    if isinstance(time, datetime.datetime):
        time = time.isoformat()

    return dotnet.lib.AF.Time.AFTime(time)


def to_af_time_span(interval: IntervalLike | None) -> dotnet.AF.Time.AFTimeSpan:
    """Convert a time interval to a AFTimeSpan value.

    Parameters
    ----------
        interval (str | datetime.timedelta | pd.Timedelta): Interval to convert to AFTimeSpan.

    Returns
    -------
        :afsdk:`AF.Time.AFTimeSpan <T_OSIsoft_AF_Time_AFTimeSpan.htm>`:
            AFTimeSpan version of interval.
    """
    if isinstance(interval, (datetime.timedelta, pd.Timedelta)):
        interval = f"{interval.total_seconds()}s"

    return dotnet.lib.AF.Time.AFTimeSpan.Parse(interval)


def timestamp_to_index(timestamp: dotnet.System.DateTime) -> datetime.datetime:
    """Convert AFTime object to datetime in local timezone.

    Parameters
    ----------
        timestamp (`System.DateTime`): Timestamp in .NET format to convert to `datetime`.

    Returns
    -------
        `datetime`: Datetime with the timezone info from :data:`PIConfig.DEFAULT_TIMEZONE <PIconnect.config.PIConfigContainer.DEFAULT_TIMEZONE>`.
    """  # noqa: E501
    local_tz = zoneinfo.ZoneInfo(PIConfig.DEFAULT_TIMEZONE)
    return (
        datetime.datetime(
            timestamp.Year,
            timestamp.Month,
            timestamp.Day,
            timestamp.Hour,
            timestamp.Minute,
            timestamp.Second,
            timestamp.Millisecond * 1000,
        )
        .replace(tzinfo=datetime.timezone.utc)
        .astimezone(local_tz)
    )



"""The Search namespace provides query based searches."""

from collections.abc import Iterable
from typing import Generic, TypeVar

from .Asset import AFAttribute, AFElement
from .Database import AFDatabase
from .EventFrame import AFEventFrame

_AFSearchable = TypeVar(
    "_AFSearchable",
)


class AFSearch(Generic[_AFSearchable]):
    """Base class for AFSearch."""

    def __init__(self, database: AFDatabase, name: str, query: str) -> None:
        self.Database = database
        self.SearchName = name
        self.TokenCollection = query

    def GetTotalCount(self) -> int:
        """Return the total count of the search results."""
        return len(self.TokenCollection)

    def FindObjects(self) -> Iterable[_AFSearchable]:
        """Return the search results."""
        obs: list[_AFSearchable] = []
        for item in obs:
            yield item


class AFAttributeSearch(AFSearch[AFAttribute]):
    """Search for AF attributes."""


class AFElementSearch(AFSearch[AFElement]):
    """Search for AF elements."""


class AFEventFrameSearch(AFSearch[AFEventFrame]):
    """Search for AF elements."""



"""Typing for AFValues and AFValue classes.

These classes are in a separate file to avoid circular imports.
"""

from typing import Any

from . import Time

_DEFAULT_TIME = Time.AFTime("MinValue")


class AFValue:
    def __init__(self, value: Any, timestamp: Time.AFTime = _DEFAULT_TIME) -> None:
        self.Value = value
        self.Timestamp = timestamp


class AFValues(list[AFValue]):
    def __init__(self):
        self.Count: int
        self.Value: list[AFValue]



"""Generics for AF collections."""

from collections.abc import Iterable, Iterator, MutableSequence
from typing import Protocol, Self, TypeVar, overload


class NamedItem(Protocol):
    """Protocol for an item with a name."""

    @property
    def name(self) -> str:
        """Return the name of the item."""
        ...


NamedItemType = TypeVar("NamedItemType", bound=NamedItem)


class NamedItemList(MutableSequence[NamedItemType]):
    """A list of items with names.

    This class provides a way to access items by index or by name.
    """

    def __init__(self, elements: MutableSequence[NamedItemType]) -> None:
        self._elements = elements

    @overload
    def __getitem__(self, index: int | str) -> NamedItemType: ...
    @overload
    def __getitem__(self, index: slice) -> Self: ...
    def __getitem__(self, index: int | str | slice) -> NamedItemType | Self:
        """Return the list item at the given index or the list item with the given name."""
        match index:
            case int():
                return self._elements[index]
            case str():
                for attr in self._elements:
                    if attr.name == index:
                        return attr
                raise KeyError(f"List item {index} not found.")
            case slice():
                return self.__class__(self._elements[index])
            case _:
                raise TypeError("Index must be an int, string or slice of int.")  # type: ignore

    def __len__(self) -> int:
        """Return the number of items in the list."""
        return len(self._elements)

    def __iter__(self) -> Iterator[NamedItemType]:
        """Return an iterator over the items in the list."""
        return iter(self._elements)

    @overload
    def __setitem__(self, index: int | str, value: NamedItemType) -> None: ...
    @overload
    def __setitem__(self, index: slice, value: Iterable[NamedItemType]) -> None: ...
    def __setitem__(
        self, index: int | str | slice, value: NamedItemType | Iterable[NamedItemType]
    ) -> None:
        """Set the list item at the given index or the list item with the given name."""
        match index:
            case int():
                self._elements[index] = value  # type: ignore
            case str():
                for i, attr in enumerate(self._elements):
                    if attr.name == index:
                        self._elements[i] = value  # type: ignore
                        return
                raise KeyError(f"List item {index} not found.")
            case slice():
                if isinstance(value, Iterable):
                    self._elements[index] = list(value)
                else:
                    raise TypeError("Value must be an iterable.")
            case _:
                raise TypeError("Index must be an int or string.")  # type: ignore

    def __delitem__(self, index: int | str | slice) -> None:
        """Delete the list item at the given index or the list item with the given name."""
        match index:
            case int():
                del self._elements[index]
            case slice():
                del self._elements[index]
            case str():
                for i, attr in enumerate(self._elements):
                    if attr.name == index:
                        del self._elements[i]
                        return
                raise KeyError(f"List item {index} not found.")
            case _:
                raise TypeError("Index must be an int or string.")  # type: ignore

    def insert(self, index: int, value: NamedItemType) -> None:
        """Insert a new item at the given index."""
        self._elements.insert(index, value)

    def append(self, value: NamedItemType) -> None:
        """Append a new item to the end of the list."""
        self._elements.append(value)

    def extend(self, values: Iterable[NamedItemType]) -> None:
        """Extend the list with a new iterable of items."""
        self._elements.extend(values)

    def __reversed__(self) -> Iterator[NamedItemType]:
        """Return a reverse iterator over the items in the list."""
        return reversed(self._elements)

    def __repr__(self) -> str:
        """Return the string representation of the list."""
        return f"{self.__class__.__qualname__}({len(self._elements)} items)"



"""Test communication with the PI AF system."""

from typing import cast

import pytest

import PIconnect as PI
from PIconnect import Asset, dotnet
from PIconnect._typing import AF

dotnet.lib.load_test_SDK()


class TestAFDatabase:
    """Test connecting to the AF database."""

    def test_connection(self):
        """Test creating a PI.AFDatabase object without arguments raises no exception."""
        PI.AFDatabase()

    def test_server_name(self):
        """Test that the server reports the same name as which was connected to."""
        AFserver = dotnet.lib.AF.PISystems().DefaultPISystem.Name
        database = dotnet.lib.AF.PISystems().DefaultPISystem.Databases.DefaultDatabase.Name
        server = PI.AFDatabase(AFserver, database)
        assert server.server_name == AFserver
        assert server.database_name == database
        assert repr(server) == "AFDatabase(\\\\{s}\\{d})".format(s=AFserver, d=database)

    def test_unknown_server_name(self):
        """Test that the server reports a warning for an unknown server."""
        AFserver_name = "__".join(list(PI.AFDatabase.servers()) + ["UnkownServerName"])
        with pytest.warns(UserWarning):
            PI.AFDatabase(server=AFserver_name)

    def test_unknown_database_name(self):
        """Test that the server reports a warning for an unknown database."""
        server = cast(AF.PISystem, PI.AFDatabase.default_server())  # type: ignore
        databases = [db.Name for db in server.Databases]
        AFdatabase_name = "__".join(databases + ["UnkownDatabaseName"])
        with pytest.warns(UserWarning):
            PI.AFDatabase(database=AFdatabase_name)


class TestDatabaseDescendants:
    """Test retrieving child elements."""

    def test_children(self):
        """Test that calling children on the database returns a dict of child elements."""
        with PI.AFDatabase() as db:
            children = db.children
            assert isinstance(children, dict)


class TestDatabaseSearch:
    """Test retrieving attributes."""

    def test_search(self):
        """Test that calling attributes on the database returns a list of attributes."""
        with pytest.warns(DeprecationWarning):
            with PI.AFDatabase() as db:
                attributes = db.search([r"", r""])
                assert isinstance(attributes, Asset.AFAttributeList)

    def test_split_element_attribute(self):
        """Test that calling attributes on the database returns a list of attributes."""
        with pytest.warns(DeprecationWarning):
            with PI.AFDatabase() as db:
                print(db.children)
                attributes = db.search(r"BaseElement|Attribute1")
                assert isinstance(attributes[0].name, str)

    def test_split_element_nested_attribute(self):
        """Test that calling attributes on the database returns a list of attributes."""
        with pytest.warns(DeprecationWarning):
            with PI.AFDatabase() as db:
                attributes = db.search(r"BaseElement|Attribute1|Attribute2")
                assert isinstance(attributes[0].name, str)



from . import Asset


class AFDatabase:
    """Mock class of the AF.AFDatabase class."""

    def __init__(self, name: str) -> None:
        self.Name = name
        self.Elements = Asset.AFElements(
            [Asset.AFElement("TestElement"), Asset.AFElement("BaseElement")]
        )
        self.Tables = Asset.AFTables([Asset.AFTable("TestTable")])



"""Data access and manipulation classes."""

import abc
import datetime
import enum
from collections.abc import Callable
from typing import Any, Concatenate, Literal, ParamSpec, TypeVar, cast

import pandas as pd  # type: ignore

import PIconnect._typing.AF as _AFtyping
from PIconnect import Time, _collections, dotnet


class BoundaryType(enum.IntEnum):
    """BoundaryType indicates how to handle the boundaries of a time range.

    Detailed information is available at
    :afsdk:`AF.Data.AFBoundaryType <T_OSIsoft_AF_Data_AFBoundaryType.htm>`.
    """

    #: The first value after the start time and the last value before the end time
    INSIDE = 0
    #: The last value before the start time and the first value after the end time
    OUTSIDE = 1
    #: Interpolate values to the specified start and end time
    INTERPOLATED = 2


class SummaryType(enum.IntFlag):
    """SummaryType indicates which types of summary should be calculated.

    `SummaryType`'s are `enum.IntFlag`'s and can be or'ed together to select
    multiple summary types. For example:

    >>> SummaryType.MINIMUM | SummaryType.MAXIMUM  # Returns minimum and maximum
    <SummaryType.MAXIMUM|MINIMUM: 12>

    Detailed information is available at
    :afsdk:`AF.Data.AFSummaryTypes <T_OSIsoft_AF_Data_AFSummaryTypes.htm>`.
    """

    #: No summary data
    NONE = 0
    #: A total over the time span
    TOTAL = 1
    #: Average value over the time span
    AVERAGE = 2
    #: The minimum value in the time span
    MINIMUM = 4
    #: The maximum value in the time span
    MAXIMUM = 8
    #: The range of the values (max-min) in the time span
    RANGE = 16
    #: The sample standard deviation of the values over the time span
    STD_DEV = 32
    #: The population standard deviation of the values over the time span
    POP_STD_DEV = 64
    #: The sum of the event count (when the calculation is event weighted).
    #: The sum of the event time duration (when the calculation is time weighted.)
    COUNT = 128
    #: The percentage of the data with a good value over the time range.
    #: Based on time for time weighted calculations,
    #: based on event count for event weigthed calculations.
    PERCENT_GOOD = 8192
    #: The total over the time span,
    #: with the unit of measurement that's associated with the input
    #: (or no units if not defined for the input).
    TOTAL_WITH_UOM = 16384
    #: A convenience to retrieve all summary types
    ALL = 24831
    #: A convenience to retrieve all summary types for non-numeric data
    ALL_FOR_NON_NUMERIC = 8320


class CalculationBasis(enum.IntEnum):
    """CalculationBasis indicates how values should be weighted over a time range.

    Detailed information is available at
    :afsdk:`AF.Data.AFCalculationBasis <T_OSIsoft_AF_Data_AFCalculationBasis.htm>`.
    """

    #: Each event is weighted according to the time over which it applies.
    TIME_WEIGHTED = 0
    #: Each event is weighted equally.
    EVENT_WEIGHTED = 1
    #: Each event is time weighted, but interpolation is always done as if it is
    #: continous data.
    TIME_WEIGHTED_CONTINUOUS = 2
    #: Each event is time weighted, but interpolation is always done as if it is
    #: discrete, stepped, data.
    TIME_WEIGHTED_DISCRETE = 3
    #: Each event is weighted equally, except data at the end of the interval is
    #: excluded.
    EVENT_WEIGHTED_EXCLUDE_MOST_RECENT = 4
    #: Each event is weighted equally, except data at the beginning of the interval
    #: is excluded.
    EVENT_WEIGHTED_EXCLUDE_EARLIEST = 5
    #: Each event is weighted equally, data at both boundaries of the interval are
    #: explicitly included.
    EVENT_WEIGHTED_INCLUDE_BOTH_ENDS = 6


class ExpressionSampleType(enum.IntEnum):
    """ExpressionSampleType indicates how expressions are evaluated over a time range.

    Detailed information is available at
    :afsdk:`AF.Data.AFSampleType <T_OSIsoft_AF_Data_AFSampleType.htm>`.
    """

    #: The expression is evaluated at each archive event.
    EXPRESSION_RECORDED_VALUES = 0
    #: The expression is evaluated at a sampling interval, passed as a separate argument.
    INTERVAL = 1


class TimestampCalculation(enum.IntEnum):
    """
    TimestampCalculation defines the timestamp returned for a given summary calculation.

    Detailed information is available at
    :afsdk:`AF.Data.AFTimeStampCalculation <T_OSIsoft_AF_Data_AFTimestampCalculation.htm>`.
    """

    #: The timestamp is the event time of the minimum or maximum for those summaries
    #: or the beginning of the interval otherwise.
    AUTO = 0
    #: The timestamp is always the beginning of the interval.
    EARLIEST_TIME = 1
    #: The timestamp is always the end of the interval.
    MOST_RECENT_TIME = 2


class RetrievalMode(enum.IntEnum):
    """RetrievalMode indicates which recorded value should be returned.

    Detailed information is available at
    :afsdk:`AF.Data.AFRetrievalMode <T_OSIsoft_AF_Data_AFRetrievalMode.htm>`.
    """

    #: Autmatic detection
    AUTO = 0
    #: At the exact time if available, else the first before the requested time
    AT_OR_BEFORE = 1
    #: The first before the requested time
    BEFORE = 6
    #: At the exact time if available, else the first after the requested time
    AT_OR_AFTER = 2
    #: The first after the requested time
    AFTER = 7
    #: At the exact time if available, else return an error
    EXACT = 4


class UpdateMode(enum.IntEnum):
    """Indicates how to treat duplicate values in the archive.

    Only used when supported by the Data Reference.

    Detailed information is available at
    :afsdk:`AF.Data.AFUpdateOption <T_OSIsoft_AF_Data_AFUpdateOption.htm>`
    """

    #: Add the value to the archive.
    #: If any values exist at the same time, will overwrite one of them and set its
    #: Substituted flag.
    REPLACE = 0
    #: Add the value to the archive. Any existing values at the same time are not overwritten.
    INSERT = 1
    #: Add the value to the archive only if no value exists at the same time.
    #: If a value already exists for that time, the passed value is ignored.
    NO_REPLACE = 2
    #: Replace an existing value in the archive at the specified time.
    #: If no existing value is found, the passed value is ignored.
    REPLACE_ONLY = 3
    #: Add the value to the archive without compression.
    #: If this value is written to the snapshot, the previous snapshot value will be written to
    #: the archive,
    #: without regard to compression settings.
    #: Note that if a subsequent snapshot value is written without the InsertNoCompression
    #: option,
    #: the value added with the InsertNoCompression option is still subject to compression.
    INSERT_NO_COMPRESSION = 5
    #: Remove the value from the archive if a value exists at the passed time.
    REMOVE = 6


class BufferMode(enum.IntEnum):
    """Indicates buffering option in updating values, when supported by the Data Reference.

    Detailed information is available at
    :afsdk:`AF.Data.AFBufferOption <T_OSIsoft_AF_Data_AFBufferOption.htm>`
    """

    #: Updating data reference values without buffer.
    DO_NOT_BUFFER = 0
    #: Try updating data reference values with buffer.
    #: If fails (e.g. data reference AFDataMethods does not support Buffering,
    #: or its Buffering system is not available),
    #: then try updating directly without buffer.
    BUFFER_IF_POSSIBLE = 1
    # Updating data reference values with buffer.
    BUFFER = 2


_DEFAULT_CALCULATION_BASIS = CalculationBasis.TIME_WEIGHTED
_DEFAULT_FILTER_EVALUATION = ExpressionSampleType.EXPRESSION_RECORDED_VALUES
_DEFAULT_TIMESTAMP_CALCULATION = TimestampCalculation.AUTO


class DataContainer(abc.ABC):
    """Abstract base class for data containers."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Return the name of the data container."""
        pass

    @property
    @abc.abstractmethod
    def stepped_data(self) -> bool:
        """Return True if the data is stepped, False if it is continuous."""
        pass

    @property
    def current_value(self) -> Any:
        """Return the current value of the attribute."""
        return self._current_value()

    @abc.abstractmethod
    def _current_value(self) -> Any:
        """Return the current value of the attribute."""
        pass

    def filtered_summaries(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        interval: Time.IntervalLike,
        filter_expression: str,
        summary_types: SummaryType,
        calculation_basis: CalculationBasis = _DEFAULT_CALCULATION_BASIS,
        filter_evaluation: ExpressionSampleType = _DEFAULT_FILTER_EVALUATION,
        filter_interval: Time.IntervalLike | None = None,
        time_type: TimestampCalculation = _DEFAULT_TIMESTAMP_CALCULATION,
    ) -> pd.DataFrame:
        """Return one or more summary values for each interval within a time range.

        Parameters
        ----------
            start_time (str or datetime): String containing the date, and possibly time,
                from which to retrieve the values. This is parsed, together
                with `end_time`, using
                :afsdk:`AF.Time.AFTimeRange <M_OSIsoft_AF_Time_AFTimeRange__ctor_1.htm>`.
            end_time (str or datetime): String containing the date, and possibly time,
                until which to retrieve values. This is parsed, together
                with `start_time`, using
                :afsdk:`AF.Time.AFTimeRange <M_OSIsoft_AF_Time_AFTimeRange__ctor_1.htm>`.
            interval (str, datetime.timedelta or pandas.Timedelta): String containing the
                interval at which to extract data. This is parsed using
                :afsdk:`AF.Time.AFTimeSpan.Parse <M_OSIsoft_AF_Time_AFTimeSpan_Parse_1.htm>`.
            filter_expression (str, optional): Defaults to ''. Query on which
                data to include in the results. See :ref:`filtering_values`
                for more information on filter queries.
            summary_types (int or PIConsts.SummaryType): Type(s) of summaries
                of the data within the requested time range.
            calculation_basis (int or PIConsts.CalculationBasis, optional):
                Event weighting within an interval. See :ref:`event_weighting`
                and :any:`CalculationBasis` for more information. Defaults to
                CalculationBasis.TIME_WEIGHTED.
            filter_evaluation (int or PIConsts.ExpressionSampleType, optional):
                Determines whether the filter is applied to the raw events in
                the database, of if it is applied to an interpolated series
                with a regular interval. Defaults to
                ExpressionSampleType.EXPRESSION_RECORDED_VALUES.
            filter_interval (str, optional): String containing the interval at
                which to extract apply the filter. This is parsed using
                :afsdk:`AF.Time.AFTimeSpan.Parse <M_OSIsoft_AF_Time_AFTimeSpan_Parse_1.htm>`.
            time_type (int or PIConsts.TimestampCalculation, optional):
                Timestamp to return for each of the requested summaries. See
                :ref:`summary_timestamps` and :any:`TimestampCalculation` for
                more information. Defaults to TimestampCalculation.AUTO.

        Returns
        -------
            pandas.DataFrame: Dataframe with the unique timestamps as row index
                and the summary name as column name.
        """
        time_range = Time.to_af_time_range(start_time, end_time)
        _interval = Time.to_af_time_span(interval)
        _filter_expression = self._normalize_filter_expression(filter_expression)
        _summary_types = dotnet.lib.AF.Data.AFSummaryTypes(int(summary_types))
        _calculation_basis = dotnet.lib.AF.Data.AFCalculationBasis(int(calculation_basis))
        _filter_evaluation = dotnet.lib.AF.Data.AFSampleType(int(filter_evaluation))
        _filter_interval = Time.to_af_time_span(filter_interval)
        _time_type = dotnet.lib.AF.Data.AFTimestampCalculation(int(time_type))
        pivalues = self._filtered_summaries(
            time_range,
            _interval,
            _filter_expression,
            _summary_types,
            _calculation_basis,
            _filter_evaluation,
            _filter_interval,
            _time_type,
        )
        df = pd.DataFrame()
        for summary in pivalues:
            key = SummaryType(int(summary.Key)).name
            timestamps, values = zip(
                *[
                    (Time.timestamp_to_index(value.Timestamp.UtcTime), value.Value)
                    for value in summary.Value
                ],
                strict=True,
            )
            df = df.join(
                pd.DataFrame(data={key: values}, index=timestamps),  # type: ignore
                how="outer",
            )
        return df

    @abc.abstractmethod
    def _filtered_summaries(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        interval: dotnet.AF.Time.AFTimeSpan,
        filter_expression: str,
        summary_types: dotnet.AF.Data.AFSummaryTypes,
        calculation_basis: dotnet.AF.Data.AFCalculationBasis,
        filter_evaluation: dotnet.AF.Data.AFSampleType,
        filter_interval: dotnet.AF.Time.AFTimeSpan,
        time_type: dotnet.AF.Data.AFTimestampCalculation,
    ) -> _AFtyping.Data.SummariesDict:
        pass

    def interpolated_value(self, time: Time.TimeLike) -> pd.Series:
        """Return a pd.Series with an interpolated value at the given time.

        Parameters
        ----------
            time (str, datetime): String containing the date, and possibly time,
                for which to retrieve the value. This is parsed, using
                :ref:`Time.to_af_time`.

        Returns
        -------
            pd.Series: A pd.Series with a single row, with the corresponding time as
                the index
        """
        _time = Time.to_af_time(time)
        pivalue = self._interpolated_value(_time)
        result = pd.Series(
            data=[pivalue.Value],
            index=[Time.timestamp_to_index(pivalue.Timestamp.UtcTime)],
            name=self.name,
        )
        result.attrs["uom"] = self.units_of_measurement
        return result

    @abc.abstractmethod
    def _interpolated_value(self, time: dotnet.AF.Time.AFTime) -> dotnet.AF.Asset.AFValue:
        pass

    def interpolated_values(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        interval: Time.IntervalLike,
        filter_expression: str = "",
    ) -> pd.Series:
        """Return a pd.Series of interpolated data.

        Data is returned between *start_time* and *end_time* at a fixed
        *interval*. All three values are parsed by AF.Time and the first two
        allow for time specification relative to "now" by use of the
        asterisk.

        *filter_expression* is an optional string to filter the returned
        values, see OSIsoft PI documentation for more information.

        The AF SDK allows for inclusion of filtered data, with filtered
        values marked as such. At this point PIconnect does not support this
        and filtered values are always left out entirely.

        Parameters
        ----------
            start_time (str or datetime): Containing the date, and possibly time,
                from which to retrieve the values. This is parsed, together
                with `end_time`, using :ref:`Time.to_af_time_range`.
            end_time (str or datetime): Containing the date, and possibly time,
                until which to retrieve values. This is parsed, together
                with `start_time`, using :ref:`Time.to_af_time_range`.
            interval (str, datetime.timedelta or pd.Timedelta): String containing the interval
                at which to extract data. This is parsed using :ref:`Time.to_af_time_span`.
            filter_expression (str, optional): Defaults to ''. Query on which
                data to include in the results. See :ref:`filtering_values`
                for more information on filter queries.

        Returns
        -------
            pd.Series: Timeseries of the values returned by the SDK
        """
        time_range = Time.to_af_time_range(start_time, end_time)
        _interval = Time.to_af_time_span(interval)
        _filter_expression = self._normalize_filter_expression(filter_expression)
        pivalues = self._interpolated_values(time_range, _interval, _filter_expression)

        timestamps: list[datetime.datetime] = []
        values: list[Any] = []
        for value in pivalues:
            timestamps.append(Time.timestamp_to_index(value.Timestamp.UtcTime))
            values.append(value.Value)
        result = pd.Series(
            data=values,
            index=timestamps,
            name=self.name,
        )
        result.attrs["uom"] = self.units_of_measurement
        return result

    @abc.abstractmethod
    def _interpolated_values(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        interval: dotnet.AF.Time.AFTimeSpan,
        filter_expression: str,
    ) -> dotnet.AF.Asset.AFValues:
        pass

    def _normalize_filter_expression(self, filter_expression: str) -> str:
        return filter_expression

    def recorded_value(
        self,
        time: Time.TimeLike,
        retrieval_mode: RetrievalMode = RetrievalMode.AUTO,
    ) -> pd.Series:
        """Return a pd.Series with the recorded value at or close to the given time.

        Parameters
        ----------
            time (str): String containing the date, and possibly time,
                for which to retrieve the value. This is parsed, using
                :afsdk:`AF.Time.AFTime <M_OSIsoft_AF_Time_AFTime__ctor_7.htm>`.
            retrieval_mode (int or :any:`PIConsts.RetrievalMode`): Flag determining
                which value to return if no value available at the exact requested
                time.

        Returns
        -------
            pd.Series: A pd.Series with a single row, with the corresponding time as
                the index
        """
        _time = Time.to_af_time(time)
        _retrieval_mode = dotnet.lib.AF.Data.AFRetrievalMode(int(retrieval_mode))
        pivalue = self._recorded_value(_time, _retrieval_mode)
        result = pd.Series(
            data=[pivalue.Value],
            index=[Time.timestamp_to_index(pivalue.Timestamp.UtcTime)],
            name=self.name,
        )
        result.attrs["uom"] = self.units_of_measurement
        return result

    @abc.abstractmethod
    def _recorded_value(
        self, time: dotnet.AF.Time.AFTime, retrieval_mode: dotnet.AF.Data.AFRetrievalMode
    ) -> dotnet.AF.Asset.AFValue:
        pass

    def recorded_values(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        boundary_type: BoundaryType = BoundaryType.INSIDE,
        filter_expression: str = "",
    ):
        """Return a pd.Series of recorded data.

        Data is returned between the given *start_time* and *end_time*,
        inclusion of the boundaries is determined by the *boundary_type*
        attribute. Both *start_time* and *end_time* are parsed by AF.Time and
        allow for time specification relative to "now" by use of the asterisk.

        By default the *boundary_type* is set to 'inside', which returns from
        the first value after *start_time* to the last value before *end_time*.
        The other options are 'outside', which returns from the last value
        before *start_time* to the first value before *end_time*, and
        'interpolate', which interpolates the first value to the given
        *start_time* and the last value to the given *end_time*.

        *filter_expression* is an optional string to filter the returned
        values, see OSIsoft PI documentation for more information.

        The AF SDK allows for inclusion of filtered data, with filtered values
        marked as such. At this point PIconnect does not support this and
        filtered values are always left out entirely.

        Parameters
        ----------
            start_time (str or datetime): Containing the date, and possibly time,
                from which to retrieve the values. This is parsed, together
                with `end_time`, using :ref:`Time.to_af_time_range`.
            end_time (str or datetime): Containing the date, and possibly time,
                until which to retrieve values. This is parsed, together
                with `start_time`, using :ref:`Time.to_af_time_range`.
            boundary_type (BoundaryType): Specification for how to handle values near the
                specified start and end time. Defaults to `BoundaryType.INSIDE`.
            filter_expression (str, optional): Defaults to ''. Query on which
                data to include in the results. See :ref:`filtering_values`
                for more information on filter queries.

        Returns
        -------
            pd.Series: Timeseries of the values returned by the SDK
        """
        time_range = Time.to_af_time_range(start_time, end_time)
        _boundary_type = dotnet.lib.AF.Data.AFBoundaryType(int(boundary_type))
        _filter_expression = self._normalize_filter_expression(filter_expression)

        pivalues = self._recorded_values(time_range, _boundary_type, _filter_expression)

        timestamps: list[datetime.datetime] = []
        values: list[Any] = []
        for value in pivalues:
            timestamps.append(Time.timestamp_to_index(value.Timestamp.UtcTime))
            values.append(value.Value)
        result = pd.Series(
            data=values,
            index=timestamps,
            name=self.name,
        )
        result.attrs["uom"] = self.units_of_measurement
        return result

    @abc.abstractmethod
    def _recorded_values(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        boundary_type: dotnet.AF.Data.AFBoundaryType,
        filter_expression: str,
    ) -> dotnet.AF.Asset.AFValues:
        """Abstract implementation for recorded values.

        The internals for retrieving recorded values from PI and PI-AF are
        different and should therefore be implemented by the respective data
        containers.
        """
        pass

    def summary(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        summary_types: SummaryType,
        calculation_basis: CalculationBasis = _DEFAULT_CALCULATION_BASIS,
        time_type: TimestampCalculation = _DEFAULT_TIMESTAMP_CALCULATION,
    ) -> pd.DataFrame:
        """Return one or more summary values over a single time range.

        Parameters
        ----------
            start_time (str or datetime): Containing the date, and possibly time,
                from which to retrieve the values. This is parsed, together
                with `end_time`, using :ref:`Time.to_af_time_range`.
            end_time (str or datetime): Containing the date, and possibly time,
                until which to retrieve values. This is parsed, together
                with `start_time`, using :ref:`Time.to_af_time_range`.
            summary_types (int or SummaryType): Type(s) of summaries
                of the data within the requested time range.
            calculation_basis (int or CalculationBasis, optional):
                Event weighting within an interval. See :ref:`event_weighting`
                and :any:`CalculationBasis` for more information. Defaults to
                CalculationBasis.TIME_WEIGHTED.
            time_type (int or TimestampCalculation, optional):
                Timestamp to return for each of the requested summaries. See
                :ref:`summary_timestamps` and :any:`TimestampCalculation` for
                more information. Defaults to TimestampCalculation.AUTO.

        Returns
        -------
            pandas.DataFrame: Dataframe with the unique timestamps as row index
                and the summary name as column name.
        """
        time_range = Time.to_af_time_range(start_time, end_time)
        _summary_types = dotnet.lib.AF.Data.AFSummaryTypes(int(summary_types))
        _calculation_basis = dotnet.lib.AF.Data.AFCalculationBasis(int(calculation_basis))
        _time_type = dotnet.lib.AF.Data.AFTimestampCalculation(int(time_type))
        pivalues = self._summary(time_range, _summary_types, _calculation_basis, _time_type)
        df = pd.DataFrame()
        for summary in pivalues:
            key = SummaryType(int(summary.Key)).name
            value = summary.Value
            timestamp = Time.timestamp_to_index(value.Timestamp.UtcTime)
            value = value.Value
            df = df.join(
                pd.DataFrame(data={key: value}, index=[timestamp]),  # type: ignore
                how="outer",
            )
        return df

    @abc.abstractmethod
    def _summary(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        summary_types: dotnet.AF.Data.AFSummaryTypes,
        calculation_basis: dotnet.AF.Data.AFCalculationBasis,
        time_type: dotnet.AF.Data.AFTimestampCalculation,
    ) -> _AFtyping.Data.SummaryDict:
        pass

    def summaries(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        interval: Time.IntervalLike,
        summary_types: SummaryType,
        calculation_basis: CalculationBasis = _DEFAULT_CALCULATION_BASIS,
        time_type: TimestampCalculation = _DEFAULT_TIMESTAMP_CALCULATION,
    ) -> pd.DataFrame:
        """Return one or more summary values for each interval within a time range.

        Parameters
        ----------
            start_time (str or datetime): Containing the date, and possibly time,
                from which to retrieve the values. This is parsed, together
                with `end_time`, using :ref:`Time.to_af_time_range`.
            end_time (str or datetime): Containing the date, and possibly time,
                until which to retrieve values. This is parsed, together
                with `start_time`, using :ref:`Time.to_af_time_range`.
            interval (str, datetime.timedelta or pd.Timedelta): String containing the interval
                at which to extract data. This is parsed using :ref:`Time.to_af_time_span`.
            summary_types (int or PIConsts.SummaryType): Type(s) of summaries
                of the data within the requested time range.
            calculation_basis (int or PIConsts.CalculationBasis, optional):
                Event weighting within an interval. See :ref:`event_weighting`
                and :any:`CalculationBasis` for more information. Defaults to
                CalculationBasis.TIME_WEIGHTED.
            time_type (int or PIConsts.TimestampCalculation, optional):
                Timestamp to return for each of the requested summaries. See
                :ref:`summary_timestamps` and :any:`TimestampCalculation` for
                more information. Defaults to TimestampCalculation.AUTO.

        Returns
        -------
            pandas.DataFrame: Dataframe with the unique timestamps as row index
                and the summary name as column name.
        """
        time_range = Time.to_af_time_range(start_time, end_time)
        _interval = Time.to_af_time_span(interval)
        _summary_types = dotnet.lib.AF.Data.AFSummaryTypes(int(summary_types))
        _calculation_basis = dotnet.lib.AF.Data.AFCalculationBasis(int(calculation_basis))
        _time_type = dotnet.lib.AF.Data.AFTimestampCalculation(int(time_type))
        pivalues = self._summaries(
            time_range, _interval, _summary_types, _calculation_basis, _time_type
        )
        df = pd.DataFrame()
        for summary in pivalues:
            key = SummaryType(int(summary.Key)).name
            timestamps, values = zip(
                *[
                    (Time.timestamp_to_index(value.Timestamp.UtcTime), value.Value)
                    for value in summary.Value
                ],
                strict=True,
            )
            df = df.join(
                pd.DataFrame(data={key: values}, index=timestamps),  # type: ignore
                how="outer",
            )
        return df

    @abc.abstractmethod
    def _summaries(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        interval: dotnet.AF.Time.AFTimeSpan,
        summary_types: dotnet.AF.Data.AFSummaryTypes,
        calculation_basis: dotnet.AF.Data.AFCalculationBasis,
        time_type: dotnet.AF.Data.AFTimestampCalculation,
    ) -> _AFtyping.Data.SummariesDict:
        pass

    @property
    @abc.abstractmethod
    def units_of_measurement(self) -> str | None:
        """Return the units of measurement of the values in the current object."""
        pass

    def update_value(
        self,
        value: Any,
        time: Time.TimeLike | None = None,
        update_mode: UpdateMode = UpdateMode.NO_REPLACE,
        buffer_mode: BufferMode = BufferMode.BUFFER_IF_POSSIBLE,
    ) -> None:
        """Update value for existing PI object.

        Parameters
        ----------
            value: value type should be in cohesion with PI object or
                it will raise PIException: [-10702] STATE Not Found
            time (datetime, optional): it is not possible to set future value,
                it raises PIException: [-11046] Target Date in Future.

        You can combine update_mode and time to change already stored value.
        """
        from . import Time as time_module

        if time is not None:
            _value = dotnet.lib.AF.Asset.AFValue(value, time_module.to_af_time(time))
        else:
            _value = dotnet.lib.AF.Asset.AFValue(value)

        _update_mode = dotnet.lib.AF.Data.AFUpdateOption(int(update_mode))
        _buffer_mode = dotnet.lib.AF.Data.AFBufferOption(int(buffer_mode))
        self._update_value(_value, _update_mode, _buffer_mode)

    @abc.abstractmethod
    def _update_value(
        self,
        value: dotnet.AF.Asset.AFValue,
        update_mode: dotnet.AF.Data.AFUpdateOption,
        buffer_mode: dotnet.AF.Data.AFBufferOption,
    ) -> None:
        pass


DataContainerType = TypeVar("DataContainerType", bound=DataContainer)
Parameters = ParamSpec("Parameters")

Align = Literal["auto", "ffill", "bfill", "nearest", "time", False]


class DataContainerCollection(_collections.NamedItemList[DataContainerType]):
    """Container for a collection of data containers."""

    @property
    def _element_type(self) -> type[DataContainerType]:
        if len(self._elements) == 0:
            return cast(type[DataContainerType], DataContainer)
        return type(self._elements[0])

    def _combine_dfs_to_df(
        self,
        func: Callable[Concatenate[DataContainerType, Parameters], pd.DataFrame | pd.Series],
        _align: Align = False,
        _add_name_to_index: bool = False,
        *args: Parameters.args,
        **kwargs: Parameters.kwargs,
    ) -> pd.DataFrame:
        """Combine the results of a function applied to each element in the collection."""

        def add_name_to_index(df: pd.DataFrame, element: DataContainerType) -> pd.DataFrame:
            if _add_name_to_index:
                return df.set_axis(  # type: ignore
                    pd.MultiIndex.from_product([[element.name], df.columns]), axis=1
                )
            return df

        def apply_func(element: DataContainerType) -> pd.DataFrame:
            result = func(element, *args, **kwargs)
            match result:
                case pd.DataFrame():
                    df = result
                case pd.Series():
                    df = result.to_frame()
            return add_name_to_index(df, element)

        def align(df: pd.DataFrame) -> pd.DataFrame:
            match _align:
                case False:
                    return df
                case "auto":
                    for col in df.columns.get_level_values(0):  # type: ignore
                        if self[str(col)].stepped_data:  # type: ignore
                            df[col] = df[col].ffill(axis=0)  # type: ignore
                        else:
                            df[col] = (
                                df[col]
                                .apply(pd.to_numeric, errors="coerce", by_row=False)  # type: ignore
                                .interpolate(method="time", axis=0)  # type: ignore
                            )
                    return df
                case "ffill":
                    return df.ffill(axis=0)  # type: ignore
                case "bfill":
                    return df.bfill(axis=0)  # type: ignore
                case "nearest":
                    return df.interpolate(method="nearest", axis=0)  # type: ignore
                case "time":
                    return df.interpolate(method="time", axis=0)  # type: ignore

        return align(
            pd.concat(
                [pd.DataFrame()] + [apply_func(e) for e in self._elements],
                axis=1,
            )
        )

    @property
    def current_value(self) -> pd.Series:
        """Return the current values of all attributes in the collection."""
        if self._elements:
            idx, value = zip(
                *[(element.name, element.current_value) for element in self._elements],
                strict=True,
            )
        else:
            idx, value = [], []
        return pd.Series(value, index=idx)

    def filtered_summaries(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        interval: Time.IntervalLike,
        filter_expression: str,
        summary_types: SummaryType,
        calculation_basis: CalculationBasis = _DEFAULT_CALCULATION_BASIS,
        filter_evaluation: ExpressionSampleType = _DEFAULT_FILTER_EVALUATION,
        filter_interval: Time.IntervalLike | None = None,
        time_type: TimestampCalculation = _DEFAULT_TIMESTAMP_CALCULATION,
        align: Align = False,
    ) -> pd.DataFrame:
        """Return one or more summary values for each interval within a time range."""
        return self._combine_dfs_to_df(
            self._element_type.filtered_summaries,
            _align=align,
            _add_name_to_index=True,
            start_time=start_time,
            end_time=end_time,
            interval=interval,
            filter_expression=filter_expression,
            summary_types=summary_types,
            calculation_basis=calculation_basis,
            filter_evaluation=filter_evaluation,
            filter_interval=filter_interval,
            time_type=time_type,
        )

    def interpolated_value(self, time: Time.TimeLike, align: Align = False) -> pd.DataFrame:
        """Return a pd.DataFrame with an interpolated value at the given time.

        .. warning::
            Relative times are evaluated for each element in the collection,
            resulting in a different time for each element. To overcome this, use
            a fixed time, for example using the datetime module:

            >>> import datetime
            >>> time = datetime.datetime.now() - datetime.timedelta(days=1)
            >>> collection.interpolated_value(time)

        Parameters
        ----------
            time (str, datetime): String containing the date, and possibly time,
                for which to retrieve the value. This is parsed, using
                :ref:`Time.to_af_time`.

        Returns
        -------
            pd.Series: A pd.Series with a single row, with the corresponding time as
                the index
        """
        return self._combine_dfs_to_df(
            self._element_type.interpolated_value, _align=align, time=time
        )

    def interpolated_values(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        interval: Time.IntervalLike,
        filter_expression: str = "",
        align: Align = False,
    ) -> pd.DataFrame:
        """Return a pd.DataFrame of interpolated data.

        Data is returned between *start_time* and *end_time* at a fixed
        *interval*. All three values are parsed by AF.Time and the first two
        allow for time specification relative to "now" by use of the
        asterisk.

        *filter_expression* is an optional string to filter the returned
        values, see OSIsoft PI documentation for more information.

        The AF SDK allows for inclusion of filtered data, with filtered
        values marked as such. At this point PIconnect does not support this
        and filtered values are always left out entirely.

        .. warning::
            Relative times are evaluated for each element in the collection,
            resulting in a different time for each element. To overcome this, use
            a fixed time, for example using the datetime module:

            >>> import datetime
            >>> time = datetime.datetime.now() - datetime.timedelta(days=1)
            >>> collection.interpolated_value(time)

        Parameters
        ----------
            start_time (str or datetime): Containing the date, and possibly time,
                from which to retrieve the values. This is parsed, together
                with `end_time`, using :ref:`Time.to_af_time_range`.
            end_time (str or datetime): Containing the date, and possibly time,
                until which to retrieve values. This is parsed, together
                with `start_time`, using :ref:`Time.to_af_time_range`.
            interval (str, datetime.timedelta or pd.Timedelta): String containing the interval
                at which to extract data. This is parsed using :ref:`Time.to_af_time_span`.
            filter_expression (str, optional): Defaults to ''. Query on which
                data to include in the results. See :ref:`filtering_values`
                for more information on filter queries.

        Returns
        -------
            pd.DataFrame: Timeseries of the values returned by the SDK
        """
        return self._combine_dfs_to_df(
            self._element_type.interpolated_values,
            _align=align,
            start_time=start_time,
            end_time=end_time,
            interval=interval,
            filter_expression=filter_expression,
        )

    def recorded_value(
        self,
        time: Time.TimeLike,
        retrieval_mode: RetrievalMode = RetrievalMode.AUTO,
        align: Align = False,
    ) -> pd.DataFrame:
        """Return a pd.Series with the recorded value at or close to the given time.

        Parameters
        ----------
            time (str): String containing the date, and possibly time,
                for which to retrieve the value. This is parsed, using
                :afsdk:`AF.Time.AFTime <M_OSIsoft_AF_Time_AFTime__ctor_7.htm>`.
            retrieval_mode (int or :any:`PIConsts.RetrievalMode`): Flag determining
                which value to return if no value available at the exact requested
                time.

        Returns
        -------
            pd.Series: A pd.Series with a single row, with the corresponding time as
                the index
        """
        return self._combine_dfs_to_df(
            self._element_type.recorded_value,
            _align=align,
            time=time,
            retrieval_mode=retrieval_mode,
        )

    def recorded_values(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        boundary_type: BoundaryType = BoundaryType.INSIDE,
        filter_expression: str = "",
        align: Align = False,
    ) -> pd.DataFrame:
        """Return a pd.Series of recorded data.

        Data is returned between the given *start_time* and *end_time*,
        inclusion of the boundaries is determined by the *boundary_type*
        attribute. Both *start_time* and *end_time* are parsed by AF.Time and
        allow for time specification relative to "now" by use of the asterisk.

        By default the *boundary_type* is set to 'inside', which returns from
        the first value after *start_time* to the last value before *end_time*.
        The other options are 'outside', which returns from the last value
        before *start_time* to the first value before *end_time*, and
        'interpolate', which interpolates the first value to the given
        *start_time* and the last value to the given *end_time*.

        *filter_expression* is an optional string to filter the returned
        values, see OSIsoft PI documentation for more information.

        The AF SDK allows for inclusion of filtered data, with filtered values
        marked as such. At this point PIconnect does not support this and
        filtered values are always left out entirely.

        Parameters
        ----------
            start_time (str or datetime): Containing the date, and possibly time,
                from which to retrieve the values. This is parsed, together
                with `end_time`, using :ref:`Time.to_af_time_range`.
            end_time (str or datetime): Containing the date, and possibly time,
                until which to retrieve values. This is parsed, together
                with `start_time`, using :ref:`Time.to_af_time_range`.
            boundary_type (BoundaryType): Specification for how to handle values near the
                specified start and end time. Defaults to `BoundaryType.INSIDE`.
            filter_expression (str, optional): Defaults to ''. Query on which
                data to include in the results. See :ref:`filtering_values`
                for more information on filter queries.

        Returns
        -------
            pd.Series: Timeseries of the values returned by the SDK
        """
        return self._combine_dfs_to_df(
            self._element_type.recorded_values,
            _align=align,
            start_time=start_time,
            end_time=end_time,
            boundary_type=boundary_type,
            filter_expression=filter_expression,
        )

    def summary(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        summary_types: SummaryType,
        calculation_basis: CalculationBasis = _DEFAULT_CALCULATION_BASIS,
        time_type: TimestampCalculation = _DEFAULT_TIMESTAMP_CALCULATION,
        align: Align = False,
    ) -> pd.DataFrame:
        """Return one or more summary values over a single time range.

        Parameters
        ----------
            start_time (str or datetime): Containing the date, and possibly time,
                from which to retrieve the values. This is parsed, together
                with `end_time`, using :ref:`Time.to_af_time_range`.
            end_time (str or datetime): Containing the date, and possibly time,
                until which to retrieve values. This is parsed, together
                with `start_time`, using :ref:`Time.to_af_time_range`.
            summary_types (int or SummaryType): Type(s) of summaries
                of the data within the requested time range.
            calculation_basis (int or CalculationBasis, optional):
                Event weighting within an interval. See :ref:`event_weighting`
                and :any:`CalculationBasis` for more information. Defaults to
                CalculationBasis.TIME_WEIGHTED.
            time_type (int or TimestampCalculation, optional):
                Timestamp to return for each of the requested summaries. See
                :ref:`summary_timestamps` and :any:`TimestampCalculation` for
                more information. Defaults to TimestampCalculation.AUTO.

        Returns
        -------
            pandas.DataFrame: Dataframe with the unique timestamps as row index
                and the summary name as column name.
        """
        return self._combine_dfs_to_df(
            self._element_type.summary,
            _align=align,
            _add_name_to_index=True,
            start_time=start_time,
            end_time=end_time,
            summary_types=summary_types,
            calculation_basis=calculation_basis,
            time_type=time_type,
        )

    def summaries(
        self,
        start_time: Time.TimeLike,
        end_time: Time.TimeLike,
        interval: Time.IntervalLike,
        summary_types: SummaryType,
        calculation_basis: CalculationBasis = _DEFAULT_CALCULATION_BASIS,
        time_type: TimestampCalculation = _DEFAULT_TIMESTAMP_CALCULATION,
        align: Align = False,
    ) -> pd.DataFrame:
        """Return one or more summary values for each interval within a time range.

        Parameters
        ----------
            start_time (str or datetime): Containing the date, and possibly time,
                from which to retrieve the values. This is parsed, together
                with `end_time`, using :ref:`Time.to_af_time_range`.
            end_time (str or datetime): Containing the date, and possibly time,
                until which to retrieve values. This is parsed, together
                with `start_time`, using :ref:`Time.to_af_time_range`.
            interval (str, datetime.timedelta or pd.Timedelta): String containing the interval
                at which to extract data. This is parsed using :ref:`Time.to_af_time_span`.
            summary_types (int or PIConsts.SummaryType): Type(s) of summaries
                of the data within the requested time range.
            calculation_basis (int or PIConsts.CalculationBasis, optional):
                Event weighting within an interval. See :ref:`event_weighting`
                and :any:`CalculationBasis` for more information. Defaults to
                CalculationBasis.TIME_WEIGHTED.
            time_type (int or PIConsts.TimestampCalculation, optional):
                Timestamp to return for each of the requested summaries. See
                :ref:`summary_timestamps` and :any:`TimestampCalculation` for
                more information. Defaults to TimestampCalculation.AUTO.

        Returns
        -------
            pandas.DataFrame: Dataframe with the unique timestamps as row index
                and the summary name as column name.
        """
        return self._combine_dfs_to_df(
            self._element_type.summaries,
            _align=align,
            _add_name_to_index=True,
            start_time=start_time,
            end_time=end_time,
            interval=interval,
            summary_types=summary_types,
            calculation_basis=calculation_basis,
            time_type=time_type,
        )



"""Mock classes for the AF namespace of the OSIsoft PI-AF SDK."""

from collections.abc import Iterator

from . import PI, Asset, Data, EventFrame, Search, Time, UnitsOfMeasure
from .Database import AFDatabase

__all__ = [
    "Asset",
    "Data",
    "EventFrame",
    "PI",
    "Search",
    "Time",
    "UnitsOfMeasure",
    "AFDatabase",
    "AFCategory",
    "PISystem",
    "PISystems",
]


class AFCategory:
    """Mock class of the AF.AFCategory class."""


class AFCategories(list[AFCategory]):
    def __init__(self, elements: list[AFCategory]) -> None:
        self.Count: int
        self._values = elements


class PISystem:
    """Mock class of the AF.PISystem class."""

    class InternalDatabases:
        """Mock class for the AF.PISystem.Databases property."""

        def __init__(self) -> None:
            self.DefaultDatabase: AFDatabase | None = AFDatabase("TestDatabase")

        def __iter__(self) -> Iterator[AFDatabase]:
            if self.DefaultDatabase is not None:
                yield from [self.DefaultDatabase]

        def __getitem__(self, name: str) -> AFDatabase | None:
            """Return the AFDatabase with the given name."""
            if self.DefaultDatabase and name == self.DefaultDatabase.Name:
                return self.DefaultDatabase

    def __init__(self, name: str) -> None:
        self.Name = name
        self.Databases = PISystem.InternalDatabases()
        self._connected = False

    def Connect(self) -> None:
        """Stub to connect to the testing system."""
        self._connected = True

    def Disconnect(self) -> None:
        """Stub to disconnect from the testing system."""
        self._connected = False


class PISystems:
    """Mock class of the AF.PISystems class."""

    Version = "0.0.0.0"

    def __init__(self) -> None:
        self.DefaultPISystem = PISystem("TestingAF")
        self.Count = 1

    def __iter__(self) -> Iterator[PISystem]:
        return (x for x in [self.DefaultPISystem])

    def __getitem__(self, name: str) -> PISystem | None:
        """Return the PISystem with the given name."""
        if name == self.DefaultPISystem.Name:
            return self.DefaultPISystem



"""Module for handling Event Frames."""

import enum
from typing import Self

from PIconnect import Asset, _collections, dotnet


class EventFrameSearchMode(enum.IntEnum):
    """EventFrameSearchMode.

    EventFrameSearchMode defines the interpretation and direction from the start time
    when searching for event frames.

    Detailed information is available at
    :afsdk:`AF.EventFrame.AFEventFrameSearchMode <T_OSIsoft_AF_EventFrame_AFEventFrameSearchMode.htm>`.
    including a graphical display of event frames that are returned for a given search
    mode.
    """  # noqa: E501

    #: Uninitialized
    NONE = 0
    #: Backward from start time, also known as starting before
    BACKWARD_FROM_START_TIME = 1
    STARTING_BEFORE = 1
    #: Forward from start time, also known as starting after
    FORWARD_FROM_START_TIME = 2
    STARTING_AFTER = 2
    #: Backward from end time, also known as ending before
    BACKWARD_FROM_END_TIME = 3
    ENDING_BEFORE = 3
    #: Forward from end time, also known as ending after
    FORWARD_FROM_END_TIME = 4
    ENDING_AFTER = 4
    #: Backward in progress, also known as starting before and in progress
    BACKWARD_IN_PROGRESS = 5
    STARTING_BEFORE_IN_PROGRESS = 5
    #: Forward in progress, also known as starting after and in progress
    FORWARD_IN_PROGRESS = 6
    STARTING_AFTER_IN_PROGRESS = 6


class AFEventFrame(Asset.AFBaseElement[dotnet.AF.EventFrame.AFEventFrame]):
    """Container for PI AF Event Frames in the database."""

    version = "0.1.0"

    @property
    def event_frame(self) -> dotnet.AF.EventFrame.AFEventFrame:
        """Return the underlying AF Event Frame object."""
        return self.element

    @property
    def parent(self) -> Self | None:
        """Return the parent element of the current event frame, or None if it has none."""
        if not self.element.Parent:
            return None
        return self.__class__(self.element.Parent)

    @property
    def children(self) -> dict[str, Self]:
        """Return a dictionary of the direct child event frames of the current event frame."""
        return {c.Name: self.__class__(c) for c in self.element.EventFrames}


class AFEventFrameList(_collections.NamedItemList[AFEventFrame]):
    """Container for a list of PIAFEventFrame objects."""

    pass



"""Mock classes for the AF.Asset module."""

from collections.abc import Iterator
from typing import cast

from PIconnect._typing import AF, System

from . import Data, Generic
from . import UnitsOfMeasure as UOM
from ._values import AFValue, AFValues

__all__ = [
    "AFAttribute",
    "AFAttributes",
    "AFBaseElement",
    "AFElement",
    "AFElements",
    "AFElementTemplate",
    "AFTable",
    "AFTables",
    "AFValue",
    "AFValues",
]


class AFAttribute:
    def __init__(
        self,
        name: str,
        parent: "AFAttribute | None" = None,
        _element: "AFBaseElement | None" = None,
    ) -> None:
        self.Attributes: AFAttributes
        self.Element: "AFBaseElement" = AFBaseElement(name) if _element is None else _element
        if parent is None:
            self.Attributes = AFAttributes(
                [
                    AFAttribute("Attribute1", parent=self, _element=self.Element),
                    AFAttribute("Attribute2", parent=self, _element=self.Element),
                ]
            )
        self.Data: Data.AFData
        self.DataReference: AFDataReference
        self.Description: str = f"Description of {name}"
        self.DefaultUOM = UOM.UOM()
        self.Name = name
        self.Parent = parent
        self.Step = False

    @staticmethod
    def GetValue() -> AFValue:
        """Stub for getting a value."""
        return AFValue(0)

    @staticmethod
    def GetPath() -> str:
        """Stub for getting the path."""
        return "Path\\to\\the|attribute"


class AFAttributes(list[AFAttribute]):
    def __init__(self, elements: list[AFAttribute]) -> None:
        self.Count: int
        self._values = elements

    def __iter__(self) -> Iterator[AFAttribute]:
        yield from self._values


class AFBaseElement:
    def __init__(self, name: str, parent: "AFElement | None" = None) -> None:
        self.Attributes = AFAttributes(
            [
                AFAttribute("Attribute1", _element=self),
                AFAttribute("Attribute2", _element=self),
            ]
        )
        self.Categories: AF.AFCategories
        self.Description: str
        self.Elements: AFElements
        if parent is None:
            self.Elements = AFElements(
                [
                    AFElement("Element1", parent=cast(AFElement, self)),
                    AFElement("Element2", parent=cast(AFElement, self)),
                    AFElement("BaseElement", parent=cast(AFElement, self)),
                ]
            )
        self.Name = name
        self.Parent = parent

    @staticmethod
    def GetPath() -> str:
        """Stub for getting the path."""
        return "Path\\to\\the|element"


class AFElement(AFBaseElement):
    """Mock class of the AF.AFElement class."""


class AFElements(list[AFElement]):
    def __init__(self, elements: list[AFElement]) -> None:
        self.Count: int
        self._values = elements

    def get_Item(self, name: str | int) -> AFElement:
        """Stub for the indexer."""
        if isinstance(name, int):
            return self._values[name]
        return AFElement(name)

    def __iter__(self) -> Iterator[AFElement]:
        yield from self._values


class AFElementTemplate:
    """Mock class of the AF.Asset.AFElementTemplate class."""


class AFEnumerationValue:
    """Mock class of the AF.Asset.AFEnumerationValue class."""

    def __init__(self, name: str, value: int) -> None:
        self.Name = name
        self.Value = value


class AFDataReference:
    from . import PI

    def __init__(
        self, name: str, attribute: AFAttribute, pi_point: PI.PIPoint | None = None
    ) -> None:
        self.Attribute = attribute
        self.Name = name
        self.PIPoint = pi_point


class AFTable:
    def __init__(self, name: str) -> None:
        self.Name = name
        self.Table: System.Data.DataTable


class AFTables(list[AFTable]):
    def __init__(self, elements: list[AFTable]) -> None:
        self.Count: int
        self._values = elements


AttributeDict = Generic.Dictionary[str, AFAttribute]



"""Mock classes for the AF.Time module."""

from PIconnect._typing import System


class AFTime:
    """Mock class of the AF.Time.AFTime class."""

    def __init__(self, time: str) -> None:
        self.UtcTime: System.DateTime

    Now: System.DateTime


class AFTimeRange:
    """Mock class of the AF.Time.AFTimeRange class."""

    def __init__(self, start_time: str, end_time: str):
        pass

    @staticmethod
    def Parse(start_time: str, end_time: str) -> "AFTimeRange":
        return AFTimeRange(start_time, end_time)


class AFTimeSpan:
    """Mock class of the AF.Time.AFTimeSpan class."""

    def __init__(self):
        pass

    @staticmethod
    def Parse(interval: str | None, /) -> "AFTimeSpan":
        """Stub for parsing strings that should return a AFTimeSpan."""
        return AFTimeSpan()



"""Type stubs for the AF SDK and dotnet libraries."""

from typing import Protocol

from . import AF, System


class AFType(Protocol):
    # Modules
    # Analysis = AF.Analysis
    Asset = AF.Asset
    # Collective = AF.Collective
    Data = AF.Data
    # Diagnostics = AF.Diagnostics
    EventFrame = AF.EventFrame
    # Modeling = AF.Modeling
    # Notification = AF.Notification
    PI = AF.PI
    Search = AF.Search
    # Support = AF.Support
    Time = AF.Time
    # UI = AF.UI
    UnitsOfMeasure = AF.UnitsOfMeasure

    # Classes
    # AFActiveDirectoryProperties = AF.AFActiveDirectoryProperties
    AFCategory = AF.AFCategory
    AFCategories = AF.AFCategories
    # AFChangedEventArgs = AF.AFChangedEventArgs
    # AFCheckoutInfo = AF.AFCheckoutInfo
    # AFClientRegistration = AF.AFClientRegistration
    # AFCollection = AF.AFCollection
    # AFCollectionList = AF.AFCollectionList
    # AFConnectionInfo = AF.AFConnectionInfo
    # AFContact = AF.AFContact
    # AFCsvColumn = AF.AFCsvColumn
    # AFCsvColumns = AF.AFCsvColumns
    AFDatabase = AF.AFDatabase
    # AFDatabases = AF.AFDatabases
    # AFErrors = AF.AFErrors
    # AFEventArgs = AF.AFEventArgs
    # AFGlobalRestorer = AF.AFGlobalRestorer
    # AFGlobalSettings = AF.AFGlobalSettings
    # AFKeyedResults = AF.AFKeyedResults
    # AFLibraries = AF.AFLibraries
    # AFLibrary = AF.AFLibrary
    # AFListResults = AF.AFListResults
    # AFNamedCollection = AF.AFNamedCollection
    # AFNamedCollectionList = AF.AFNamedCollectionList
    # AFNameSubstitution = AF.AFNameSubstitution
    # AFObject = AF.AFObject
    # AFOidcIdentity = AF.AFOidcIdentity
    # AFPlugin = AF.AFPlugin
    # AFPlugins = AF.AFPlugins
    # AFProgressEventArgs = AF.AFProgressEventArgs
    # AFProvider = AF.AFProvider
    # AFRole = AF.AFRole
    # AFSDKExtension = AF.AFSDKExtension
    # AFSecurity = AF.AFSecurity
    # AFSecurityIdentities = AF.AFSecurityIdentities
    # AFSecurityIdentity = AF.AFSecurityIdentity
    # AFSecurityMapping = AF.AFSecurityMapping
    # AFSecurityMappings = AF.AFSecurityMappings
    # AFSecurityRightsExtension = AF.AFSecurityRightsExtension
    # NumericStringComparer = AF.NumericStringComparer
    PISystem = AF.PISystem
    PISystems = AF.PISystems
    # UniversalComparer = AF.UniversalComparer


class SystemType(Protocol):
    # Modules
    Data = System.Data
    Net = System.Net
    Security = System.Security

    # Classes
    DateTime = System.DateTime
    Exception = System.Exception
    TimeSpan = System.TimeSpan


AF_SDK_VERSION = "2.7_compatible"

__all__ = ["AF", "AF_SDK_VERSION", "AFType", "System"]



"""Mock for System.* classes."""

from typing import Protocol

from . import Data, Net, Security

__all__ = [
    "Data",
    "Exception",
    "Net",
    "Security",
    "TimeSpan",
]


class Exception:
    pass


class TimeSpan:
    def __init__(self, /, hours: int, minutes: int, seconds: int) -> None:
        self.Hours = hours
        self.Minutes = minutes
        self.Seconds = seconds


class DateTime(Protocol):
    """Mock for System.DateTime."""

    Year: int
    Month: int
    Day: int
    Hour: int
    Minute: int
    Second: int
    Millisecond: int



"""Mock classes for the System.Data module."""

from collections.abc import Iterator
from typing import Any

__all__ = [
    "DataRow",
    "DataRowCollection",
    "DataColumn",
    "DataColumnCollection",
    "DataTable",
]


class DataRow:
    __fields: dict[str, Any]

    def __getitem__(self, key: str) -> Any:
        return self.__fields[key]


class DataRowCollection:
    rows: list[DataRow]

    def __iter__(self) -> Iterator[DataRow]:
        yield from self.rows


class DataColumn:
    def __init__(self, name: str) -> None:
        self.ColumnName = name


class DataColumnCollection:
    columns: list[DataColumn]

    def __iter__(self) -> Iterator[DataColumn]:
        yield from self.columns


class DataTable:
    Name: str
    Rows: list[DataRow]
    Columns: list[DataColumn]



"""Mock classes for the AF.EventFrame namespace of the OSIsoft PI-AF SDK."""

import enum
from collections.abc import Iterable

from PIconnect._typing import AF
from PIconnect._typing.AF import Asset, Time


class AFEventFrameSearchMode(enum.IntEnum):
    None_ = 0
    BackwardFromStartTime = 1
    ForwardFromStartTime = 2
    BackwardFromEndTime = 3
    ForwardFromEndTime = 4
    BackwardInProgress = 5
    ForwardInProgress = 6


class AFEventFrame(Asset.AFBaseElement):
    """Mock class of the AF.EventFrame.AFEventFrame class."""

    def __init__(self, name: str, parent: "AFEventFrame | None" = None) -> None:
        self.Name = name
        self.Parent = parent
        self.EventFrames: AFEventFrames

    @staticmethod
    def FindEventFrames(
        database: "AF.AFDatabase",
        search_root: "AFEventFrame | None",
        start_time: Time.AFTime,
        start_index: int,
        max_count: int,
        search_mode: AFEventFrameSearchMode,
        name_filter: str | None = None,
        referenced_element_name_filter: str | None = None,
        element_category: "AF.AFCategory | None" = None,
        element_template: Asset.AFElementTemplate | None = None,
        search_full_hierarchy: bool = False,
        /,
    ) -> Iterable["AFEventFrame"]:
        """Mock method to find event frames."""
        _config = {
            "database": database,
            "search_root": search_root,
            "start_time": start_time,
            "start_index": start_index,
            "max_count": max_count,
            "search_mode": search_mode,
            "name_filter": name_filter,
            "referenced_element_name_filter": referenced_element_name_filter,
            "element_category": element_category,
            "element_template": element_template,
            "search_full_hierarchy": search_full_hierarchy,
        }
        return []


class AFEventFrames(list[AFEventFrame]):
    def __init__(self, elements: list[AFEventFrame]) -> None:
        self.Count: int
        self._values = elements



"""Mock classes for the System.Net module."""

from .Security import SecureString


class NetworkCredential:
    def __init__(
        self, username: str, password: SecureString, domain: str | None = None
    ) -> None:
        self.UserName = username
        self.Password = password
        self.Domain = domain



"""Mirror of the OSISoft.AF.Asset namespace."""

import dataclasses
from typing import Generic, Self, TypeVar, overload

import pandas as pd  # type: ignore

import PIconnect._typing.AF as _AFtyping
from PIconnect import PI, Data, _collections, dotnet

__all__ = [
    "AFDataReference",
    "AFAttribute",
    "AFAttributeList",
]

T = TypeVar("T")
ElementType = TypeVar("ElementType", bound=dotnet.AF.Asset.AFBaseElement)


@dataclasses.dataclass
class AFDataReference:
    """Reference to the data source of an AF attribute."""

    data_reference: dotnet.AF.Asset.AFDataReference

    @property
    def attribute(self) -> "AFAttribute":
        """Return the attribute associated with the data reference."""
        return AFAttribute(self.data_reference.Attribute)

    @property
    def name(self) -> str:
        """Return the name of the data reference."""
        return self.data_reference.Name

    @property
    def pi_point(self) -> PI.PIPoint | None:
        """Return the PI Point associated with the data reference, if any."""
        if self.data_reference.PIPoint is not None:
            return PI.PIPoint(self.data_reference.PIPoint)


class AFEnumerationValue:
    """Representation of an AF enumeration value."""

    def __init__(self, value: dotnet.AF.Asset.AFEnumerationValue) -> None:
        self._value = value

    def __str__(self) -> str:
        """Return the string representation of the enumeration value."""
        return self._value.Name

    def __int__(self) -> int:
        """Return the integer representation of the enumeration value."""
        return self._value.Value

    def __repr__(self):
        """Return the string representation of the enumeration value."""
        return f"{self.__class__.__qualname__}({self._value.Name})"

    @property
    def name(self) -> str:
        """Return the name of the enumeration value."""
        return self._value.Name

    @property
    def value(self) -> int:
        """Return the integer value of the enumeration value."""
        return self._value.Value

    @overload
    @staticmethod
    def wrap_enumeration_value(
        value: dotnet.AF.Asset.AFEnumerationValue,
    ) -> "AFEnumerationValue": ...
    @overload
    @staticmethod
    def wrap_enumeration_value(
        value: T,
    ) -> T: ...
    @staticmethod
    def wrap_enumeration_value(
        value: T | dotnet.AF.Asset.AFEnumerationValue,
    ) -> "T | AFEnumerationValue":
        """Wrap the value in an AFEnumerationValue if it is an enumeration value."""
        if isinstance(value, dotnet.lib.AF.Asset.AFEnumerationValue):
            return AFEnumerationValue(value)
        return value


class AFAttribute(Data.DataContainer):
    """Representation of an AF attribute."""

    def __init__(self, attribute: dotnet.AF.Asset.AFAttribute) -> None:
        super().__init__()
        self.attribute = attribute

    def __repr__(self):
        """Return the string representation of the current attribute."""
        description = ", ".join([x for x in [self.name, self.description] if x])
        value = " ".join(
            [str(x) for x in [self.current_value, self.units_of_measurement] if x]
        )
        return f"{self.__class__.__qualname__}({description}; Current Value: {value})"

    @property
    def stepped_data(self) -> bool:
        """Return True if the attribute is a stepped data type."""
        return self.attribute.Step

    @property
    def element(self) -> dotnet.AF.Asset.AFBaseElement:
        """Return the element to which the attribute belongs."""
        return self.attribute.Element

    @property
    def parent(self) -> Self | None:
        """Return the parent attribute of the current attribute, or None if it has none."""
        if not self.attribute.Parent:
            return None
        return self.__class__(self.attribute.Parent)

    @property
    def children(self) -> dict[str, Self]:
        """Return a dictionary of the direct child attributes of the current attribute."""
        return {a.Name: self.__class__(a) for a in self.attribute.Attributes}

    @property
    def path(self) -> str:
        """Return the path of the attribute."""
        return self.attribute.GetPath()

    @property
    def name(self) -> str:
        """Return the name of the attribute."""
        return self.path.split("\\")[-1]

    @property
    def data_reference(self) -> AFDataReference:
        """Return the data reference of the attribute."""
        return AFDataReference(self.attribute.DataReference)

    @property
    def description(self) -> str:
        """Return the description of the attribute."""
        return self.attribute.Description

    @property
    def units_of_measurement(self) -> str:
        """Return the units of measurement of the attribute."""
        return str(self.attribute.DefaultUOM or "")

    def _normalize_filter_expression(self, filter_expression: str) -> str:
        return super()._normalize_filter_expression(
            filter_expression.replace("%attribute%", f"'{self.attribute.Name}'")
        )

    def _current_value(self) -> object:
        """Return the current value of the attribute."""
        return AFEnumerationValue.wrap_enumeration_value(self.attribute.GetValue().Value)

    def _filtered_summaries(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        interval: dotnet.AF.Time.AFTimeSpan,
        filter_expression: str,
        summary_types: dotnet.AF.Data.AFSummaryTypes,
        calculation_basis: dotnet.AF.Data.AFCalculationBasis,
        filter_evaluation: dotnet.AF.Data.AFSampleType,
        filter_interval: dotnet.AF.Time.AFTimeSpan,
        time_type: dotnet.AF.Data.AFTimestampCalculation,
    ) -> _AFtyping.Data.SummariesDict:
        return self.attribute.Data.FilteredSummaries(
            time_range,
            interval,
            filter_expression,
            summary_types,
            calculation_basis,
            filter_evaluation,
            filter_interval,
            time_type,
        )

    def _interpolated_value(self, time: dotnet.AF.Time.AFTime):
        """Return a single value for this PI Point."""
        return self.attribute.Data.InterpolatedValue(time, self.attribute.DefaultUOM)

    def _recorded_value(
        self, time: dotnet.AF.Time.AFTime, retrieval_mode: dotnet.AF.Data.AFRetrievalMode
    ) -> dotnet.AF.Asset.AFValue:
        """Return a single value for this PI Point."""
        return self.attribute.Data.RecordedValue(
            time, retrieval_mode, self.attribute.DefaultUOM
        )

    def _recorded_values(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        boundary_type: dotnet.AF.Data.AFBoundaryType,
        filter_expression: str,
    ) -> dotnet.AF.Asset.AFValues:
        include_filtered_values = False
        return self.attribute.Data.RecordedValues(
            time_range,
            boundary_type,
            self.attribute.DefaultUOM,
            filter_expression,
            include_filtered_values,
        )

    def _interpolated_values(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        interval: dotnet.AF.Time.AFTimeSpan,
        filter_expression: str,
    ) -> dotnet.AF.Asset.AFValues:
        """Query the pi af attribute, internal implementation."""
        include_filtered_values = False
        return self.attribute.Data.InterpolatedValues(
            time_range,
            interval,
            self.attribute.DefaultUOM,
            filter_expression,
            include_filtered_values,
        )

    def _summaries(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        interval: dotnet.AF.Time.AFTimeSpan,
        summary_types: dotnet.AF.Data.AFSummaryTypes,
        calculation_basis: dotnet.AF.Data.AFCalculationBasis,
        time_type: dotnet.AF.Data.AFTimestampCalculation,
    ) -> _AFtyping.Data.SummariesDict:
        return self.attribute.Data.Summaries(
            time_range, interval, summary_types, calculation_basis, time_type
        )

    def _summary(
        self,
        time_range: dotnet.AF.Time.AFTimeRange,
        summary_types: dotnet.AF.Data.AFSummaryTypes,
        calculation_basis: dotnet.AF.Data.AFCalculationBasis,
        time_type: dotnet.AF.Data.AFTimestampCalculation,
    ) -> _AFtyping.Data.SummaryDict:
        return self.attribute.Data.Summary(
            time_range, summary_types, calculation_basis, time_type
        )

    def _update_value(
        self,
        value: dotnet.AF.Asset.AFValue,
        update_mode: dotnet.AF.Data.AFUpdateOption,
        buffer_mode: dotnet.AF.Data.AFBufferOption,
    ) -> None:
        return self.attribute.Data.UpdateValue(
            value,
            update_mode,
            buffer_mode,
        )


class AFAttributeList(Data.DataContainerCollection[AFAttribute]):
    """A list of AF attributes."""

    pass


class AFBaseElement(Generic[ElementType]):
    """Container for PI AF elements in the database."""

    version = "0.1.0"

    def __init__(self, element: ElementType) -> None:
        self.element = element

    def __repr__(self) -> str:
        """Return the string representation of the element."""
        return f"{self.__class__.__qualname__}({self.name})"

    @property
    def name(self) -> str:
        """Return the name of the current element."""
        return self.element.Name

    @property
    def attributes(self) -> dict[str, AFAttribute]:
        """Return a dictionary of the attributes of the current element."""
        return {a.Name: AFAttribute(a) for a in self.element.Attributes}

    @property
    def categories(self) -> dotnet.AF.AFCategories:
        """Return the categories of the current element."""
        return self.element.Categories

    @property
    def description(self) -> str:
        """Return the description of the current element."""
        return self.element.Description

    @property
    def path(self) -> str:
        """Return the path of the current element."""
        return self.element.GetPath()


class AFElement(AFBaseElement[dotnet.AF.Asset.AFElement]):
    """Container for PI AF elements in the database."""

    version = "0.1.0"

    @property
    def parent(self) -> Self | None:
        """Return the parent element of the current element, or None if it has none."""
        if not self.element.Parent:
            return None
        return self.__class__(self.element.Parent)

    @property
    def children(self) -> dict[str, Self]:
        """Return a dictionary of the direct child elements of the current element."""
        return {c.Name: self.__class__(c) for c in self.element.Elements}

    def descendant(self, path: str) -> Self:
        """Return a descendant of the current element from an exact path."""
        return self.__class__(self.element.Elements.get_Item(path))


class AFElementList(_collections.NamedItemList[AFElement]):
    """Container for a list of PIAFElement objects."""

    pass


class AFTable:
    """Container for PI AF Tables in the database."""

    def __init__(self, table: dotnet.AF.Asset.AFTable) -> None:
        self._table = table

    @property
    def columns(self) -> list[str]:
        """Return the names of the columns in the table."""
        return [col.ColumnName for col in self._table.Table.Columns]

    @property
    def _rows(self) -> list[dotnet.System.Data.DataRow]:
        return self._table.Table.Rows

    @property
    def name(self) -> str:
        """Return the name of the table."""
        return self._table.Name

    @property
    def shape(self) -> tuple[int, int]:
        """Return the shape of the table."""
        return (len(self._rows), len(self.columns))

    @property
    def data(self) -> pd.DataFrame:
        """Return the data in the table as a pandas DataFrame."""
        return pd.DataFrame([{col: row[col] for col in self.columns} for row in self._rows])
